#========================================================================================
# package      DE_EPAGES::Core::API::Archive
# state        public
#----------------------------------------------------------------------------------------
# description  Provides functionality for unpacking archives and prevent unpacking malicious files.
#               Some checks are done by the function ExtractFilesFromZipArchive
#========================================================================================
package DE_EPAGES::Core::API::Archive;
use base Exporter;
use strict;

our @EXPORT_OK = qw(
    ExtractFilesFromZipArchive
);

use Archive::Zip qw( :ERROR_CODES :CONSTANTS );
use DE_EPAGES::Core::API::Error qw( Error );
use DE_EPAGES::Core::API::Warning qw( Warning );
use DE_EPAGES::Core::API::File qw( DeleteFile IsImageFile IsPDFFile );
use DE_EPAGES::Core::API::TestParameter qw( TestString );
use DE_EPAGES::Core::API::Unicode qw ( Unicode2Filename );

use constant {
    CONVERSION_FACTOR_KB => 1024,
    DEFAULT_MAX_ARCHIVEMEMBER_FILESIZE_KB => 102400
};

sub ErrorDefinition {
    return {
        READ_ZIP_ERROR => 'Invalid archive found. (Filename: #FileName)',
        FileNotFound => 'Archive file not found. (Filename: #FileName)',
        POTENTIAL_MALICIOUS_ARCHIVE => 'Archive potentialy malicious.(ArchiveName: #ArchiveName, Filename: #FileName, UncompressedFileSize in kB: #UncompressedFileSize)',
    };
}

#========================================================================================
# function     ExtractFilesFromZipArchive
# state        public
#----------------------------------------------------------------------------------------
# syntax       ExtractFilesFromZipArchive($ArchiveFile, $TargetFolder, $hParams);
#----------------------------------------------------------------------------------------
# description  Extract files from given archive to the TargetFolder
#               The archive
#----------------------------------------------------------------------------------------
# input        $ArchiveFile | zip archive full path and filename | string
# input        $TargetFolder | target folder for extracting zip archive | string
# input        $hParams | arguments for this function:
#               <ul>
#                   <li>DeleteAfterExtraction - (optional, default=1) delete archive after extracting - boolean
#                   <li>AllowOnlyImagesAndPDF - (optional, default=1) allow only image and pdf files - boolean
#                        (If 1 only image and pdf files are extracted)
#                   <li>FixPathErrors - (optional, default=0) path errors are corrected otherwise ignored 
#                       (If 1 filepath with paths above $TargetFolder are shortened to $TargetFolder) - boolean
#                   <li>WarningOnFileError - (optional, default=1) warn if a path or filesize error was detected - boolean
#                       (for warn if we have potential malicious files)        
#                   <li>MaxContentLength - (optional, default=102400 kB) max uncompressed size in kB of a member in the archive - boolean
#                       (files with uncompressed size > MaxContentLength are ignored )        
#               </ul>|reference to hash
#========================================================================================
sub ExtractFilesFromZipArchive {
    my ($ArchiveFile, $TargetFolder, $hParams) = @_;
    TestString( 'FileName', $ArchiveFile );
    TestString( 'TargetFolder', $TargetFolder );

    # some defaults
    $hParams->{'AllowOnlyImagesAndPDF'} //= 1;
    $hParams->{'DeleteAfterExtraction'} //= 1;
    $hParams->{'FixPathErrors'}         //= 0;
    $hParams->{'WarningOnFileError'}    //= 1;
    $hParams->{'MaxContentLength'}      //= DEFAULT_MAX_ARCHIVEMEMBER_FILESIZE_KB;

    Error( 'FileNotFound', { FileName => $ArchiveFile } ) unless -f $ArchiveFile;

    my $Zip = Archive::Zip->new();
    Error( 'READ_ZIP_ERROR', { FileName => $ArchiveFile } ) unless($Zip->read($ArchiveFile) == AZ_OK);

    foreach my $Member ($Zip->members()) {
        my $UncompressedFileSize = $Member->uncompressedSize() / CONVERSION_FACTOR_KB;
        my $TooLarge;

        if ($hParams->{'MaxContentLength'}){
            # prevent zip bombs by checking the uncompressed site by MaxContentLength
            # see DEFAULT_MAX_ARCHIVEMEMBER_FILESIZE_KB
            $TooLarge = 1 if $UncompressedFileSize > $hParams->{'MaxContentLength'};  
        }

        # prevent using ../ in path or / at start (UNITY-4585 security issue)
        # otherwise we can store files outside the $TargetFolder
        my $FilePathAndName = $Member->fileName;
        if($TooLarge or $FilePathAndName =~ m|(\Q../\E)+|g or $FilePathAndName =~ m|^/|){
            # warn if we have problems with the path in the archive
            Warning('POTENTIAL_MALICIOUS_ARCHIVE',{'FileName' => $FilePathAndName,
                                                   'ArchiveName' => $ArchiveFile,
                                                   'UncompressedFileSize' => $UncompressedFileSize })
                                                 if $hParams->{'WarningOnFileError'};

            # we do not accept file size problems                                                 
            next if $TooLarge;

            # we ignore the path problems if wanted...
            next unless $hParams->{'FixPathErrors'};

            # replace ../ and / at the beginning
            $FilePathAndName =~ s|(\Q../\E)+||g;
            $FilePathAndName =~ s|^/||;

        }            

        # type check: we allow only images and pdf (UNITY-4585 security issue)
        if($hParams->{'AllowOnlyImagesAndPDF'}){
            next unless (IsImageFile($FilePathAndName) or IsPDFFile($FilePathAndName));
        }

        #-- $FilePathAndName can include a path, since archives can have folders. The only allowed char in sub Unicode2Filename is '.'.
        #-- So, also allow '/'. We would also include '\', but we don't run on windows anymore, so forget it!
        $Zip->extractMember($Member->fileName, $TargetFolder . Unicode2Filename($FilePathAndName, '/'));
    }

    DeleteFile($ArchiveFile) if $hParams->{'DeleteAfterExtraction'};

}

1;
