#==============================================================================
# package      DE_EPAGES::Customer::UI::Shop
# state        private
#==============================================================================
package DE_EPAGES::Customer::UI::Shop;
use base DE_EPAGES::Presentation::UI::Login;

use strict;
use DateTime;
use HTTP::Status qw( HTTP_NOT_FOUND );
use Scalar::Util qw ( blessed );
use DE_EPAGES::Address::API::Address qw( CheckZipcode );
use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError);
use DE_EPAGES::Core::API::Log qw ( LogDebug LogError );
use DE_EPAGES::Core::API::Password qw ( CryptPassword );
use DE_EPAGES::Core::API::PerlTools qw ( CreateSecureCode AddDuration ElementOf isValid );
use DE_EPAGES::Core::API::Url qw ( BuildParameterString );
use DE_EPAGES::Core::API::Warning qw ( Warning );
use DE_EPAGES::Core::API::Scramble qw ( Scramble );
use DE_EPAGES::Customer::API::Shop qw ( GetNextCustomerNo GetCustomerFolder InsertCustomerForUser );
use DE_EPAGES::Customer::API::User qw ( UpdateCustomerTaxModel );
use DE_EPAGES::Customer::API::Constants qw ( REGISTRATION_FIELDS_MAX_LENGTH SPAM_FIELD_VALUES );
use DE_EPAGES::Customer::UI::CustomerAttribute;
use DE_EPAGES::Database::API::Connection qw ( GetCurrentDBHandle );
use DE_EPAGES::Object::API::Factory qw ( ExistsObject LoadClassByAlias Transaction );
use DE_EPAGES::Object::API::Table::Country qw ( GetInfoCountry );
use DE_EPAGES::Presentation::API::User qw ( ExistsUserByEMail CreateUser );
use DE_EPAGES::Search::API::Object::PagerSearch qw ( ReplaceWildcards ReplaceWildcardsWithoutPosOne );
use DE_EPAGES::Search::API::SearchDefinition qw (GetSearchDefinition);
use DE_EPAGES::Shop::API::Address qw ( CheckStringVATID CheckVATID UpdateVATID );
use DE_EPAGES::Shop::API::Shop qw ( GetNextNo );
use DE_EPAGES::ShopMailType::API::ShopMailType qw( GetShopMailType );
use DE_EPAGES::Trigger::API::Trigger qw( TriggerHook );


sub ErrorDefinition {
    return {
        'LoginRequired' => 'Access to shop #ShopAlias is restricted to registered customers.'
    }
}

#==============================================================================
# Event handlers
#==============================================================================
sub ViewRegistration {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    my $RegistrationSet = $Shop->get('RegistrationSet');

    # get registration set as plain
    my %Vars;
    foreach my $Info (@$RegistrationSet) {
        my $Alias = $Info->{'Attribute'}->alias;
        $Vars{$Alias.'IsVisible'}   = $Info->{'IsVisible'};
        $Vars{$Alias.'IsMandatory'} = $Info->{'IsMandatory'};
    }
    return \%Vars;
}

sub MobileSF_ViewRegistration {
    my $self = shift;
    my ($Servlet) = @_;

    # EPG-23968: set status code 404 for the mobile registration page to remove it
    # from search engines
    $Servlet->response->code( HTTP_NOT_FOUND );
    return;
}

#========================================================================================
# function     SendCustomerWelcomeMail
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Sent = SendCustomerWelcomeMail($Servlet, $User, $Customer, $Password);
#----------------------------------------------------------------------------------------
# description  Sends the welcome e-mail for new customers if the e-mail event
#               CustomerWelcome is active or if the double-opt-in process is active.
#----------------------------------------------------------------------------------------
# input        $Servlet | servlet object | object
# input        $User | user object | object
# input        $Customer | customer object | object
# input        $Password | new password | string
# return       $Sent | true if the e-mail has been sent | boolean
#========================================================================================
sub SendCustomerWelcomeMail {
    my $self = shift;
    my ($Servlet, $User, $Customer, $Password) = @_;

    my $Shop = $Servlet->object;
    my $ViewActionPrefix = $Shop->isUnity() ? 'Unity' : '';

    #send order confirmation mail to customer
    my $MailType = GetShopMailType( $Shop, 'CustomerWelcome' );
    return 0 unless $MailType->get('IsActive') || $MailType->get('IsAlwaysActive');

    #Create Header
    my %MailParams = (
        'TO'          => $User->get('EMail'),
        'HTMLAllowed' => $User->get('IsHtmlEMailAllowed'),
    );

    my $ConfirmURL = join '/',
        $Shop->get('WebServerScriptUrlSSL'),
        $Servlet->locale,
        '?'.BuildParameterString({
            ViewObjectID => $User->id,
            ViewAction => $ViewActionPrefix . 'ViewMyAccount',
            ChangeObjectID => $User->getSite->id,
            ChangeAction => 'ConfirmUserAccount',
            Alias => Scramble($User->alias, ''),
            EMail => Scramble($User->get('EMail'), ''),
            SecurityCode => CreateSecureCode($User->alias.$User->get('EMail')),
            ErrorAction =>  $ViewActionPrefix . 'ViewMyAccount',
            ErrorObjectID => $User->id,
        });

    my %Vars = (
        'Password'          => $Password,
        'Customer'          => $Customer->tleHash,
        'User'              => $User->tleHash,
        'Site'              => $Shop->tleHash,
        'ConfirmURL'        => $ConfirmURL,
    );
    my %RegionalParams = (
        Locale      => $Servlet->params('Locale'),
        Language    => $Servlet->params('Language'),
    );

    # don't event try to send an e-mail if the merchant doesn't have a sender e-mail
    my $Sent = 0;
    my $From = $MailType->from;
    if( defined $From && $From =~ /\S/ ) {
        eval {
            $MailType->send( $Shop, \%MailParams, \%RegionalParams, \%Vars );
            $Sent = 1;
        };
        if (ExistsError()) {
            my $Error = GetError();
            LogError($Error);
        }
    }

    return $Sent;
}

sub checkRegistrationSet {
    my $self = shift;
    my ($Servlet, $Shop, $AddressPrefix, $RegistrationSet) = @_;

    TriggerHook('UI_RegistrationAddressValidation', {'Servlet' => $Servlet});

    $AddressPrefix = '' unless defined $AddressPrefix;
    $RegistrationSet = $AddressPrefix unless defined $RegistrationSet;

    my $ClassAddress = LoadClassByAlias('Address');
    my $MandRegistrationSetName    = 'Is'.$RegistrationSet.'Mandatory';
    my $VisibleRegistrationSetName = 'Is'.$RegistrationSet.'Visible';
    my $Form = $Servlet->form;
    my %Address = ();

    my $USCountry = 0;
    if ($Form->value('CountryID')) {
        $USCountry = 1 if GetInfoCountry($Form->value('CountryID'))->{'Code2'} eq 'US';
    }

    foreach my $Set (grep {$_->{$VisibleRegistrationSetName} || ($USCountry && $_->{'Attribute'}->alias eq 'State')} @{$Shop->get('RegistrationSet')}) {

        # remove these lines when attributes are removed from all non-member registration sets
        my $Attribute = $Set->{'Attribute'};
        my $AttributeName = $Attribute->alias;
        next unless $ClassAddress->existsAttribute( $AttributeName ) or $RegistrationSet eq '';

        # get form value eg: 'ShippingAddressCountry'
        my $FieldName = $AddressPrefix eq '' ? $AttributeName : $AddressPrefix.'.'.$AttributeName;
        my $hFieldDef = { Name => $FieldName, Form => $AddressPrefix eq '' ? 'Address' : $AddressPrefix};
        $hFieldDef->{Mandatory} = 1 if $Set->{$MandRegistrationSetName} || ($USCountry && $AttributeName eq 'State');

        my $Type = $Attribute->get('Type');
        if( $Type eq 'String' ) {
            $hFieldDef->{Type} = 'string';
            $hFieldDef->{MaxValue} = $Attribute->get('Length');
        }
        elsif( $Type eq 'File' ) {
            $hFieldDef->{Type} = 'file';
        }
        elsif( $Type eq 'EMail' ) {
            $hFieldDef->{Type} = 'email_address';
        }
        elsif( $Attribute->get('IsObject') ) {
            $hFieldDef->{Type} = 'integer';
            $hFieldDef->{MinValue} = 1;
        }
        else { # Integer, Float, DateTime, Money
            $hFieldDef->{Type} = 'reg_'. lc $Type;
        }
        my $OrgValue = $Form->value( $FieldName );
        $Address{$AttributeName} = $Form->field( $Shop, $OrgValue, $hFieldDef );
    }
    return \%Address;
}

#========================================================================================
# function     checkRegistration
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->checkRegistration($Servlet);
#----------------------------------------------------------------------------------------
# description  Evaluates registration information such as Login and Password. Adds
#               form errors either if the Login is missing or if it exists already or if
#               the password exists or if both passwords don't match. Checks if all data
#               which were set mandatory due to the registration set have been entered.
#               Used at storefront only.
#----------------------------------------------------------------------------------------
# input        $Servlet | current servlet (contains form data) | object
# return       $hValues | attribute name => value hash | ref.hash
#========================================================================================
sub checkRegistration {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;
    my $Form = $Servlet->form;

    # skip form errors, to show all form errors together
    my $hUser = $Form->form($Shop, 'SaveRegistration', 1);

    if (not defined $hUser->{'UserName'} or $hUser->{'UserName'} !~ /\S/) {
        $Form->addFormError({
            'Reason'=> 'NO_USER_NAME',
            'Form'  => 'SaveRegistration',
            'Name'  => 'UserName',
        });
    }
    elsif (ExistsUserByEMail($Shop, $hUser->{'UserName'})) {
        $Form->addFormError({
            'Reason'=> 'LOGIN_EXISTS',
            'Form'  => 'SaveRegistration',
            'Name'  => 'UserName',
        });
    }
    # compare passwords
    $hUser->{$_} = '' foreach grep { not defined $hUser->{$_} } (qw (PasswordConf Password));
    if ($hUser->{'Password'} !~ /\S/) {
        $Form->addFormError({
            'Reason'=> 'NO_USER_PWD',
            'Form'  => 'SaveRegistration',
            'Name'  => 'Password',
        });
    }
    elsif ($hUser->{'Password'} ne $hUser->{'PasswordConf'}) {
        $Form->addFormError({
            'Reason'=> 'PWD_NOT_SAME',
            'Form'  => 'SaveRegistration',
            'Name'  => 'PasswordConf',
        });
        $Form->addFormError({
            'Reason'=> 'PWD_NOT_SAME',
            'Form'  => 'SaveRegistration',
            'Name'  => 'Password',
        });
    }

    # test registration set
    $self->checkRegistrationSet($Servlet, $Shop);

    # test the Privacy Policy checkbox
    if( $Shop->get('AcceptPrivacyPolicy') ) {
        unless( $Form->value('AcceptPrivacyPolicy') ) {
            $Form->addFormError({
                'Reason'=> 'STRING_EMPTY',
                'Form'  => 'SaveRegistration',
                'Name'  => 'AcceptPrivacyPolicy',
            });
        }
    }

    my %AttributeValues;
    $AttributeValues{'User'}     = $self->attributeValues($Servlet, $Shop, LoadClassByAlias('User'));
    $AttributeValues{'Address'}  = $self->attributeValues($Servlet, $Shop, LoadClassByAlias('Address'));
    $AttributeValues{'Customer'} = $self->attributeValues($Servlet, $Shop, $Shop->get('CustomerClass'));

    my $VATID = $AttributeValues{'Address'}{'Values'}->{'VATID'};
    # test valid EU-VATID
    if ($VATID) {
        if (!CheckStringVATID($VATID)) {
            $Form->addFormError({
                'Reason'=> 'VATIDIsNotValid',
                'Form'  => 'SaveRegistration',
                'Name'  => 'VATID',
            });
        }
        else {
            my $CountryID = $Form->value('CountryID');
            $VATID = UpdateVATID($VATID, $CountryID, $Shop);
            my $check = CheckVATID($VATID, $CountryID, $Shop);
            $Servlet->vars('VATID', $VATID);
            if ( 1 == $check ) {
                $AttributeValues{'Address'}{'Values'}->{'VATID'} = $VATID;
            }
            else {
                $Form->addFormError({
                    'Reason'=> $check ? 'VATIDIsNotFromSameCountry' : 'VATIDIsNotValid',
                    'Form'  => 'SaveRegistration',
                    'Name'  => 'VATID',
                });
            }
        }
    }

    # test validation of zipcode
    if ($Form->value('Zipcode')) {
        my $Zipcode = $AttributeValues{'Address'}{'Values'}->{'Zipcode'};
        my $CountryID = $Form->value('CountryID');
        $Form->addFormError({
            'Reason'=> 'ZIPCODE_NOT_VALID',
            'Form'  => 'SaveRegistration',
            'Name'  => 'Zipcode',
        }) unless CheckZipcode($Zipcode, $CountryID);
    }

    return \%AttributeValues;
}

#========================================================================================
# function     saveRegistration
# state        public
#----------------------------------------------------------------------------------------
# syntax       ($User, $Customer, $Password, $NewTaxModel) = $self->saveRegistration($Servlet);
# syntax       ($User, $Customer, $Password, $NewTaxModel) = $self->saveRegistration($Servlet, $hShippingAddress);
#----------------------------------------------------------------------------------------
# description  Creates new user, sets user billing address (mandatory) and shipping
#               address (optional). Logs the user in and sends off the welcome email.
#               Used at storefront only.
#----------------------------------------------------------------------------------------
# input        $Servlet | current servlet (contains form data) | object
# input        $hAttributeValues | validated user, customer and adress values.
#                                   see also return of checkRegistration
#                           | ref.hash
# return       $User | new user | DE_EPAGES::Permission::API::Object::User
# return       $User | new customer | DE_EPAGES::Customer::API::Object::Customer
# return       $Password | password | string
# return       $NewTaxModel | tax model (0=net, 1=gross) | int
#========================================================================================
sub saveRegistration {
    my $self = shift;
    my ($Servlet, $hAttributeValues) = @_;
    my $Shop = $Servlet->object;
    my $Form = $Servlet->form;
    my $IsDoubleOptInRegistrationDisabled = $Shop->get('IsDoubleOptInRegistrationDisabled');

    my $hAddressValues = $hAttributeValues->{'Address'}{'Values'};

    my $hUser = $Form->form($Shop, 'SaveRegistration', 1);

    $hUser->{'EMail'} = $hUser->{'UserName'};

    if (defined $hAddressValues->{'DisplayName'}) {
        $hUser->{'Name'}  = $hAddressValues->{'DisplayName'};
    } else {
        my $FirstName = $hAddressValues->{'FirstName'};
        my $LastName  = $hAddressValues->{'LastName'};
        $hUser->{'Name'} = $FirstName if defined $FirstName;
        $hUser->{'Name'} .= ' '.  $LastName if defined $LastName;
        if (defined $hUser->{'Name'}) {
            $hUser->{'Name'} =~ s/^\s+//g;
            $hUser->{'Name'} =~ s/\s+$//g;
        }
    }
    $hUser->{'Name'}  = $hUser->{'UserName'} unless defined $hUser->{'Name'};
    $hUser->{'LocaleID'} = $Servlet->locale;
    $hUser->{'CurrencyID'} = $Servlet->currencyID;
    $hUser->{'IsActivated'} = $IsDoubleOptInRegistrationDisabled ? 1 : 0;
    $hUser->{'IsEmailVerified'} = $IsDoubleOptInRegistrationDisabled ? 1 : 0;
    my $Password = $hUser->{'Password'};
    $hUser->{'Password'} = eval { CryptPassword($hUser->{'Password'}); };
    if (ExistsError()) {
        my $Error = GetError();
        Error($Error) unless $Error->code eq 'NativeDie' and $Error->shortMessage eq 'Wide character in crypt';
        $Servlet->form->executeFormError({
            'Reason'=> 'PASSWORD_WRONG',
            'Form'  => 'SaveRegistration',
            'Name'  => 'Password',
        }) ;
    }

    delete $hUser->{'UserName'};
    delete $hUser->{'PasswordConf'};

    my $User;
    my $Customer;
    Transaction( sub {
        # insert user
        $User = CreateUser($Shop, { %{$hAttributeValues->{'User'}{'Values'}}, %$hUser });
        if ( exists $hAttributeValues->{'Adresses'} &&
            exists $hAttributeValues->{'Adresses'}->{'Shipping'} )
        {
            my $Shipping = $hAttributeValues->{'Adresses'}->{'Shipping'};
            $Shipping->{'Alias'} = $User->folder('Addresses')->nextAlias('ShippingAddress %s'),
            my $UserShippingAddress = $User->folder('Addresses')->insertChild('Address', $Shipping);
            $User->set({'ShippingAddress' => $UserShippingAddress});
        }
        my $UserBillingAddress = $User->folder('Addresses')->insertChild('Address', {
            %{$hAttributeValues->{'Address'}{'Values'}},
            'Alias' => $User->folder('Addresses')->nextAlias('BillingAddress %s'),
        });
        $User->set({'BillingAddress' => $UserBillingAddress});

        # insert customer
        my $hhCustomerAttributes = $self->attributeValues($Servlet, $Shop, $Shop->get('CustomerClass'));
        $Customer = InsertCustomerForUser($User, $Shop, $hAttributeValues->{'Customer'}{'Values'});
        $Customer->set( $hhCustomerAttributes->{'Values'} );
        $Customer->set( $hhCustomerAttributes->{'LocalizedValues'}{$_}, $_ ) foreach keys %{$hhCustomerAttributes->{'LocalizedValues'}};

    });
    my $NewTaxModel = UpdateCustomerTaxModel($User, $Shop);

    return ($User, $Customer, $Password, $NewTaxModel);
}

# handles the form "Become a customer"
# registers a new customer in the storefront
sub SaveRegistration {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object;
    my $IsUnityShop = $Shop->isUnity();
    my $ViewActionPrefix = $IsUnityShop ? 'Unity' : '';
    my $DefaultViewAction = $IsUnityShop ? 'UnityViewRedirectToStorefront' : 'View';
    my $Form = $Servlet->form;

    $self->GoogleCaptcha($Servlet);

    # EPG-32268: don't allow non-SSL registrations if SSL is available
    $Servlet->testSecureConnection();

    # EPG-36714
    $Servlet->testPostRequest();

    #spam check - check if field senderName has value,
    #this field is only for spam protection it is hidden for the customer
    if($Form->value('SenderName')) {
        $Servlet->vars('ViewAction', $DefaultViewAction);
        return;
    }

    # EPC-396 Change newsletter form to become uninteresting for spammers
    $self->CheckRegistrationFields($Servlet);

    # bugfix 3541: don't allow registration if shop is restricted to registered customers
    if( $Shop->get('LoginRequired') ) {
        Warning( 'LoginRequired', { 'ShopAlias' => $Shop->alias } );
        $Servlet->vars('ViewAction', $DefaultViewAction);
        return;
    }

    if ($IsUnityShop) {
        my $FormValues = $Form->values;
        $FormValues->{'EMail'} = $Form->value('UserName');
        $Form->setValues($FormValues);
    }

    my $hAttributeValues = $self->checkRegistration($Servlet);
    # newsletter is not subscribed in checkout
    $Servlet->params('SubscribedInCheckout', 0);
    TriggerHook('UI_BeforeSaveRegistration', {'Servlet' => $Servlet});
    $Form->executeFormError;

    my ($User, $Customer, $Password, $NewTaxModel) = $self->saveRegistration($Servlet, $hAttributeValues);
    my $MailSent = $self->SendCustomerWelcomeMail($Servlet, $User, $Customer, $Password);


    # show information about changed tax model only if the shop is not VATreleased
    $Servlet->params('TaxModelChanged', 1) if defined $NewTaxModel and not $Shop->get('VATreleased');

    my $IsDoubleOptInRegistrationDisabled = $Shop->get('IsDoubleOptInRegistrationDisabled');
    if( $IsDoubleOptInRegistrationDisabled ) {
        $self->LoginUser($Servlet, $User);
        $Servlet->vars('EMAIL_VERIFIED', 1);
        $Servlet->vars('ViewAction', $ViewActionPrefix . 'ViewMyAccount');
    }
    else {
        $Servlet->vars('ViewAction', $ViewActionPrefix . 'ViewDoubleOptInMessage');
        if( $MailSent ) {
            $Servlet->params('VERIFY_WELCOME_EMAIL', 1);
        }
        else {
            $Servlet->params('SEND_WELCOME_MAIL_FAILED',1);
        }
    }

    TriggerHook('UI_AfterSaveRegistration', {'Servlet' => $Servlet, 'User' => $User, 'Customer' => $Customer});
    return $User;
}

sub _EmptyAttributeHash ($$$) {
    my ($Shop, $hAttributes, $AttributeID) = @_;

    $hAttributes->{$AttributeID} = {
        'Attribute' => $Shop->loadObject($AttributeID, 'Attribute', { AllowSystemSite => 1 }),
        'IsVisible' => 0,
        'IsMandatory' => 0,
        'IsBillingAddressVisible' => 0,
        'IsBillingAddressMandatory' => 0,
        'IsShippingAddressVisible' => 0,
        'IsShippingAddressMandatory' => 0,
    };
}

sub SaveRegistrationSet {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    my $Values = $Servlet->form->formValues;
    my %Attributes; # keys are IDs values IsVisible,...
    foreach my $AttributeID (map { /^IsVisible_(.*)/x } grep { /^IsVisible_/x } keys %$Values) {
        _EmptyAttributeHash($Shop, \%Attributes, $AttributeID) unless exists $Attributes{$AttributeID};
        $Attributes{$AttributeID}->{'IsVisible'} = 1;
    }
    foreach my $AttributeID (map { /^IsMandatory_(.*)/x } grep { /^IsMandatory_/x } keys %$Values) {
        _EmptyAttributeHash($Shop, \%Attributes, $AttributeID) unless exists $Attributes{$AttributeID};
        $Attributes{$AttributeID}->{'IsMandatory'} = 1;
    }
    foreach my $AttributeID (map { /^IsBillingAddressVisible_(.*)/x } grep { /^IsBillingAddressVisible_/x } keys %$Values) {
        _EmptyAttributeHash($Shop, \%Attributes, $AttributeID) unless exists $Attributes{$AttributeID};
        $Attributes{$AttributeID}->{'IsBillingAddressVisible'} = 1;
    }
    foreach my $AttributeID (map { /^IsBillingAddressMandatory_(.*)/x } grep { /^IsBillingAddressMandatory_/x } keys %$Values) {
        _EmptyAttributeHash($Shop, \%Attributes, $AttributeID) unless exists $Attributes{$AttributeID};
        $Attributes{$AttributeID}->{'IsBillingAddressMandatory'} = 1;
    }
    foreach my $AttributeID (map { /^IsShippingAddressVisible_(.*)/x } grep { /^IsShippingAddressVisible_/x } keys %$Values) {
        _EmptyAttributeHash($Shop, \%Attributes, $AttributeID) unless exists $Attributes{$AttributeID};
        $Attributes{$AttributeID}->{'IsShippingAddressVisible'} = 1;
    }
    foreach my $AttributeID (map { /^IsShippingAddressMandatory_(.*)/x } grep { /^IsShippingAddressMandatory_/x } keys %$Values) {
        _EmptyAttributeHash($Shop, \%Attributes, $AttributeID) unless exists $Attributes{$AttributeID};
        $Attributes{$AttributeID}->{'IsShippingAddressMandatory'} = 1;
    }
    my @RegistrationSet = values %Attributes;
    $Shop->set({'RegistrationSet'=>\@RegistrationSet});
    return;
}

sub UnitySaveRegistrationSet {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    my $Values = $Servlet->form->formValues;
    my %Attributes; # keys are IDs values IsVisible,...
    # set visibility for identical for registration, billing address and shipping address
    foreach my $AttributeID (map { /^IsVisible_(.*)/x } grep { /^IsVisible_/x } keys %$Values) {
        _EmptyAttributeHash($Shop, \%Attributes, $AttributeID) unless exists $Attributes{$AttributeID};
        $Attributes{$AttributeID}->{'IsVisible'} = 1;
        $Attributes{$AttributeID}->{'IsBillingAddressVisible'} = 1;
        $Attributes{$AttributeID}->{'IsShippingAddressVisible'} = 1;
    }
    # set mandatory fields for identical for registration and billing address
    foreach my $AttributeID (map { /^IsMandatory_(.*)/x } grep { /^IsMandatory_/x } keys %$Values) {
        _EmptyAttributeHash($Shop, \%Attributes, $AttributeID) unless exists $Attributes{$AttributeID};
        $Attributes{$AttributeID}->{'IsMandatory'} = 1;
        $Attributes{$AttributeID}->{'IsBillingAddressMandatory'} = 1;
    }
    # allow different mandatory fields for the shipping address
    foreach my $AttributeID (map { /^IsShippingAddressMandatory_(.*)/x } grep { /^IsShippingAddressMandatory_/x } keys %$Values) {
        _EmptyAttributeHash($Shop, \%Attributes, $AttributeID) unless exists $Attributes{$AttributeID};
        $Attributes{$AttributeID}->{'IsShippingAddressMandatory'} = 1;
    }
    my @UnityRegistrationSet = values %Attributes;
    $Shop->set({'UnityRegistrationSet'=>\@UnityRegistrationSet});
    return;
}

sub UnitySaveCustomerGroup {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    if (defined $Servlet->params('NewAlias') and $Servlet->params('NewAlias') =~ /\S/) {
        my $NewForm = $Servlet->form->form($Shop, 'UnityNewCustomerGroup');
        my %Info;
        $Info{$_}= $NewForm->{'New'.$_} foreach map { s/^New// && $_} grep { /^New/ } keys %$NewForm;
        $Info{'Shop'} = $Shop;

        # workaround for current simple pricelist handling with predefined customergroup and one default pricelist
        if (!$Shop->folder('Groups')->findChild('NewCustomer') && ($Info{'Alias'} ne 'NewCustomer')) {
            my %InfoNewCustomer;
            $InfoNewCustomer{'Alias'} = 'NewCustomer';
            $InfoNewCustomer{'Shop'} = $Shop;
            $InfoNewCustomer{'IsNewMemberDefault'} = 1;
            $InfoNewCustomer{'IsNonMemberDefault'} = 1;
            $Shop->folder('Groups')->insertChild('CustomerGroup', \%InfoNewCustomer);
        }

        # Form error if group already exists or group name is NewCustomer
        if ($Info{'Alias'} eq 'NewCustomer') {
            $Servlet->form->addFormError({
                'Reason' => 'AliasUsedBySystem',
                'Name'   => 'NewAlias',
                'Form'   => 'UnityNewCustomerGroup',
            });
        } elsif ($Shop->folder('Groups')->findChild($Info{'Alias'})) {
            $Servlet->form->addFormError({
                'Reason' => 'CustomerGroupExists',
                'Name'   => 'NewAlias',
                'Form'   => 'UnityNewCustomerGroup',
            });
        }
        $Servlet->form->executeFormError();

        eval {
            Transaction( sub {
                $Shop->folder('Groups')->insertChild('CustomerGroup', \%Info);
            });
        };
        if (ExistsError()) {
            my $Error = GetError();
            if ($Error->code  eq 'FeatureLimitExceeded'){
                $Servlet->vars('FeatureMaxValueReached', $Error->vars);
            } else {
                Error($Error);
            }
        }
        return;
    }
    return;
}

sub SaveCustomerGroups {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    my $Form = $Servlet->form->form($Shop, 'SaveCustomerGroups');
    my $NewDefault = $Form->{'IsNewMemberDefault'};
    delete $Form->{'IsNewMemberDefault'};
    my $NonMemberDefault = $Form->{'IsNonMemberDefault'};
    delete $Form->{'IsNonMemberDefault'};

    # delete checked objects and return
    if ($Servlet->params('delete')) {
        return if ! $Servlet->params('ListObjectID');

        $Servlet->params('DELETE',1);

        if ($Servlet->params('ASSIGN_CUSTOMER_GROUP')) {
            $self->_ReAssignCustomerGroup($Servlet);
        } else {
            my $hValues = $Servlet->form->form($Shop, 'ListedObjects');
            my @ListObjects = map  { $Shop->loadObject($_->{'ListObjectID'}, 'CustomerGroup') }
                                @{$hValues->{'ListObjectIDs'}};
            my $CustomerGroupsCount = $Shop->featureCurrentValue( 'CustomerGroups' );
            if ($CustomerGroupsCount > scalar (@ListObjects)) {
                my $CustomersCount = map  { @{$_->get('Customers')} } @ListObjects;
                if ($CustomersCount) {
                    $Servlet->vars('DeleteObjects', [ map { $_->tleHash } @ListObjects]);
                    $Servlet->vars('ASSIGN_CUSTOMERS_TO_GROUP',$CustomersCount);
                    return;
                }
            }
        }
        $self->DeleteListedObjects($Servlet);
        _SetDefaultGroup($Shop,$NewDefault, $NonMemberDefault);
    }


    if ($Servlet->params('save')) {
        # insert new if alias \S
        if (defined $Servlet->params('NewAlias') and $Servlet->params('NewAlias') =~ /\S/) {
            my $NewForm = $Servlet->form->form($Shop, 'NewCustomerGroup');
            my %Info;
            $Info{$_}= $NewForm->{'New'.$_} foreach map { s/^New// && $_} grep { /^New/ } keys %$NewForm;
            $Info{'Shop'} = $Shop;

            if ($Shop->folder('Groups')->findChild($Info{'Alias'})) {
                $Servlet->vars('DOUBLE_NEW_OBJECT',1);
                $Servlet->vars('NewAlias',$Info{'Alias'});
                return;
            }

            eval {
                Transaction( sub {
                    my $NewGroup = $Shop->folder('Groups')->insertChild('CustomerGroup', \%Info);

                    foreach my $LanguageEntry (@{$NewForm->{'Languages'}}) {
                        my %_Info;
                        $_Info{$_} = $LanguageEntry->{'New'.$_} foreach map { substr($_, 3) } keys %$LanguageEntry;
                        delete $_Info{'LanguageID'};
                        $NewGroup->set(\%_Info, $LanguageEntry->{'NewLanguageID'});
                    }
                    _SetDefaultGroup($Shop ,$NewDefault, $NonMemberDefault);
                });
            };
            if (ExistsError()) {
                my $Error = GetError();
                if ($Error->code  eq 'FeatureLimitExceeded'){
                    $Servlet->vars('FeatureMaxValueReached', $Error->vars);
                } else {
                    Error($Error);
                }
            }
        }
        $Servlet->form->executeFormError;
        foreach my $LanguageEntry (@{$Form->{'Languages'}}) {
            my %Info;
            $Info{$_} = $LanguageEntry->{$_} foreach keys %$LanguageEntry;
            delete $Info{'LanguageID'};
            delete $Info{'GroupID'};
            $Shop->loadObject($LanguageEntry->{'GroupID'}, 'CustomerGroup')->set(\%Info, $LanguageEntry->{'LanguageID'});
        }
        _SetDefaultGroup($Shop,$NewDefault, $NonMemberDefault);
        return;

    }
    return;
}

sub _SetDefaultGroup {
    my ($Shop, $NewDefault, $NonMemberDefault) = @_;

    if (defined $NewDefault and ExistsObject($NewDefault)){
        my $oNewDefault = $Shop->loadObject($NewDefault, 'CustomerGroup');
        die 'invalid Customergroup:'.$oNewDefault->id.' used in Shop:'.$Shop->id unless $oNewDefault->get('Parent')->id == $Shop->child('Groups')->id;
        $Shop->set({'NewMemberCustomerGroup' => $oNewDefault});
    }

    if (defined $NonMemberDefault and ExistsObject($NonMemberDefault)){
        my $oNonMemberDefault = $Shop->loadObject($NonMemberDefault, 'CustomerGroup');
        die 'invalid Customergroup:'.$oNonMemberDefault->id.' used in Shop:'.$Shop->id unless $oNonMemberDefault->parent->id == $Shop->child('Groups')->id;
        $Shop->set({'NonMemberCustomerGroup'=>$oNonMemberDefault})
    }

    return;
}

sub ViewNewCustomer {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    my %Defaults;
    foreach my $Class ($Shop->get('CustomerClass'), LoadClassByAlias('Address'), LoadClassByAlias('User')) {
        $Servlet->vars('Class', $Class);
        my $Defaults = $self->ViewNewObject($Servlet);
        if ($Class->alias eq 'User') {
            $Defaults{'User'.$_} = $Defaults->{$_} foreach keys %$Defaults;
        } else {
            $Defaults{$_} = $Defaults->{$_} foreach keys %$Defaults;
        }
    }
    return \%Defaults;
}

sub NewCustomer {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    # check address fields before create customer
    my $AddressClass = LoadClassByAlias( 'Address' );
    $self->attributeValues( $Servlet, $Shop, $AddressClass );

    my $Customer;
    Transaction( sub {
        # insert customer
        my $CustomerFolder = GetCustomerFolder($Shop);
        my $CustomerClass = $Shop->get('CustomerClass');
        $Customer = $self->NewObject( $Servlet, $CustomerFolder, $CustomerClass, { 'Alias' => GetNextCustomerNo($Shop) } );

        # insert address
        my $BillingAddress = $self->NewObject( $Servlet, $Customer, $AddressClass );
        $Customer->set({'BillingAddress' => $BillingAddress});
    });

    # set customer to default next view object
    return $Customer;
}

sub ViewCustomers {
    my $self = shift;
    my ($Servlet) = @_;

    return $self->SearchCustomers($Servlet, 1);
}

sub UnityViewCustomers {
    my $self = shift;
    my ($Servlet) = @_;

    # trigger hook to allow the adding of context sensitive apps
    TriggerHook('UI_Shop_UnityViewCustomers', { 'Servlet' => $Servlet });

    return $self->ViewCustomers($Servlet);
}

sub UnitySearchCustomers {
    my $self = shift;
    my ($Servlet) = @_;

    # trigger hook to allow the adding of context sensitive apps
    TriggerHook('UI_Shop_UnitySearchCustomers', { 'Servlet' => $Servlet });

    return $self->SearchCustomers($Servlet);
}

sub SaveCustomerAttributes {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    my $Alias = $Servlet->form->formValue('Alias');
    if( $Alias ) {
        DE_EPAGES::Customer::UI::CustomerAttribute->NewCustomerAttribute( $Servlet );
    }

    my $CustomerClass = $Shop->get('CustomerClass');
    my @VisibleIDs = $Servlet->form->value('ListIsVisible');

    $Servlet->form->executeFormError;

    # convert array to hash for faster access
    my %VisibleIDs = map { $_ => 1 } @VisibleIDs;

    Transaction( sub {
        my $ahListItems = $self->listItems( $Servlet, $Shop, 'CustomerAttribute' );
        foreach my $hListItem (@$ahListItems) {
            my $Position = $hListItem->{'Position'};
            my $Attribute = $hListItem->{'Object'};
            my $IsVisible = $VisibleIDs{$Attribute->id} ? 1 : 0;

            # security check - verify that the object is a child of given parent
            my $AttributeClass = $Attribute->get('AttributeClass');
            if( $AttributeClass->id != $CustomerClass->id ) {
                Error( 'WrongAttributeClass', {
                    AttributeClassID => $AttributeClass->id,
                    CustomerClassID => $CustomerClass->id,
                } );
            }

            $Attribute->set( { Position => $Position, IsVisible => $IsVisible } );
        }
    });

    $CustomerClass->folder('Attributes')->renumberChildren;
    return;
}

sub DeleteCustomerAttributes {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    $self->SUPER::DeleteListedObjects( $Servlet );

    my $CustomerClass = $Shop->get('CustomerClass');
    $CustomerClass->folder('Attributes')->renumberChildren;
}

sub SearchCustomers {
    my $self = shift;
    my ($Servlet) = @_;
    my ($IsViewAction) = @_;
    my $Shop = $Servlet->object;
    my $SearchForm = $Servlet->form->form($Shop, 'SearchCustomers', $IsViewAction ? 1 : 0);

    delete $SearchForm->{$_} foreach grep { not defined $SearchForm->{$_} } keys %$SearchForm;

    my $OrderDesc = 1;
    if (defined($SearchForm->{'Search'})) {
        my $CustomOrderDesc = $Servlet->params('OrderDesc');
        $OrderDesc = $CustomOrderDesc if defined $CustomOrderDesc;
    }

    my $SearchName = $SearchForm->{'Search'} || 'Login';
    my $SearchDefinition = GetSearchDefinition(LoadClassByAlias('Customer'), $SearchName);

    my %Options = (
        'Servlet' => $Servlet,
        'Object'  => $Shop,
        'SearchDefinition' => $SearchDefinition,
        'Parameters' => {
            'Shop.ID'    => $Shop->id,
            'LanguageID' => $Servlet->languageID,
        },
        # set default values (no custom values!)
        'OrderBy'    => 'CreationDate',
        'OrderDesc'  => 1,
        'UriParameters'  => {
            'Search' => $SearchName,
        },
    );
    $Servlet->vars('OrderDesc',$OrderDesc);

    delete $SearchForm->{'Search'};
    foreach my $Parameter (keys %$SearchForm) {
        if ( $Parameter eq 'ProductAliasString' ) {

            ## UNITY-4143
            #replace '*' with '%' and add % to the end for (Alias/Productnumber)
            $Options{'Parameters'}{$Parameter} = ReplaceWildcardsWithoutPosOne($SearchForm->{$Parameter});
        } elsif ($Parameter =~ /String/) {
            # replace * with %
            $Options{'Parameters'}{$Parameter} = ReplaceWildcards($SearchForm->{$Parameter});
        } elsif ($Parameter eq 'CustomerGroupID') {
            $Options{'Parameters'}{'CustomerGroup.ID'} = $SearchForm->{$Parameter};

        } elsif ($Parameter eq 'RevenueTo'){
            #add 1 day to the searchdate
            $Options{'Parameters'}{$Parameter} = AddDuration($SearchForm->{$Parameter}->clone(), 'days', 1) if blessed( $SearchForm->{$Parameter} );
        } elsif($Parameter eq 'ProductTo'){
            #add 1 day to the searchdate
            $Options{'Parameters'}{$Parameter} = AddDuration($SearchForm->{$Parameter}->clone(), 'days', 1) if blessed( $SearchForm->{$Parameter} );
        } elsif ($Parameter eq 'CountryID') {
            $Options{'Parameters'}{'Country.ID'} = $SearchForm->{$Parameter};

        } else{
            $Options{'Parameters'}{$Parameter} = $SearchForm->{$Parameter};
        }

        $Options{'UriParameters'}{$Parameter} = $Servlet->form->value($Parameter);
    }
    # add revenue dates
    if (exists $Options{'Parameters'}{'Revenue'}) {
        if (not exists $Options{'Parameters'}{'RevenueFrom'}) {
            $Options{'Parameters'}{'RevenueFrom'} = DateTime->new('year'=>1900, 'month'=>1,'day'=>1);
        }
        if (not exists $Options{'Parameters'}{'RevenueTo'}) {
            $Options{'Parameters'}{'RevenueTo'} = GetCurrentDBHandle()->currentDateTime;
        }
    }
    # add product order dates
    if (exists $Options{'Parameters'}{'ProductAliasString'}) {
        if (not exists $Options{'Parameters'}{'ProductFrom'}) {
            $Options{'Parameters'}{'ProductFrom'} = DateTime->new('year'=>1900, 'month'=>1,'day'=>1);
        }
        if (not exists $Options{'Parameters'}{'ProductTo'}) {
            $Options{'Parameters'}{'ProductTo'} = GetCurrentDBHandle()->currentDateTime;
        }
    }
    LogDebug('SearchForm', $SearchForm);
    LogDebug('Parameters', $Options{'Parameters'});

    my $Search = DE_EPAGES::Search::API::Object::PagerSearch->new(\%Options);
    return $Search->page;
}

sub AssignCustomerGroup {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    my $hValues = $Servlet->form->form($Shop, 'ListedObjects');
    my @CustomerIDs = map { $_->{'ListObjectID'} } @{$hValues->{'ListObjectIDs'}};
    my @Customers = @{ $Shop->loadObjects( \@CustomerIDs, 'Customer') };

    if (not $Servlet->params('ASSIGN_CUSTOMER_GROUP') or not $Servlet->form->formValue('CustomerGroup')) {
        return $Servlet->vars('ASSIGN_CUSTOMER_GROUP', 0) unless scalar @Customers;
        $Servlet->vars('SelectedObjects', [map {$_->tleHash} @Customers]);
        $Servlet->vars('ASSIGN_CUSTOMER_GROUP', 1);
        return;
    }

    my $Group = $Shop->loadObject($Servlet->form->formValue('CustomerGroup'), 'CustomerGroup');
    $_->set({'CustomerGroup', $Group}) foreach @Customers;

    return;
}

sub _ReAssignCustomerGroup {
    my $self      = shift;
    my ($Servlet) = @_;

    my $Form = $Servlet->form;
    my $Shop = $Servlet->object;

    # select the replacement customer group (or undef)
    my $NewGroup;
    if ($Form->formValue('AssignCustomers')) {
        my $NewGroupID = $Servlet->form->formValue('CustomerGroup');
        $NewGroup = $Shop->loadObject($NewGroupID, 'CustomerGroup') if $NewGroupID;
    }

    # load the customer groups that are to be deleted
    my $hValues = $Form->form($Shop, 'ListedObjects');
    my @CustomerGroupIDs = map { $_->{'ListObjectID'} } @{$hValues->{'ListObjectIDs'}};
    my @CustomerGroups = @{ $Shop->loadObjects( \@CustomerGroupIDs, 'CustomerGroup') };

    # change the customer group of all customers of the deleted groups
    foreach my $CustomerGroup( @CustomerGroups ) {
        $_->set({'CustomerGroup', $NewGroup}) foreach @{ $CustomerGroup->get('Customers') };
    }

    return;
}

sub ExportCustomers {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object;
    my $hValues = $Servlet->form->form($Shop, 'ListedObjects');
    my @CustomerIDs = map { $_->{'ListObjectID'} } @{$hValues->{'ListObjectIDs'}};
    my @Customers = @{ $Shop->loadObjects( \@CustomerIDs, 'Customer') };
    return unless scalar @Customers;

    $Servlet->vars('SelectedObjects', [map {$_->tleHash} @Customers]);
    $Servlet->vars('CSVExportImportType', 'Customer');
    $Servlet->vars('EXPORT_SELECTED_OBJECTS', 1);
    $Servlet->vars('ViewActionAbort', $Servlet->params('ViewAction'));
    $Servlet->vars('ObjectIDAbort', $Shop->id);
    $Servlet->vars('ViewAction', 'MBO-ViewCSVExportImportCustomer');
    $Servlet->vars('ObjectID', $Shop->getSite->id);

    return;
}

sub ViewRegistrationSSL {
    my $self = shift;
    my ($Servlet) = @_;

    $Servlet->redirectToSecureConnection($Servlet->object->getSite->get('WebServerScriptUrlSSL'));

    return;
}

sub UnityViewNewCustomer {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object;

    $Servlet->params('Alias', GetNextNo($Shop, 'Customer',undef,0));
    $Servlet->params('IsNewCustomer', 1);

    # add form data for the maxlength validation
    my $CustomerClass = LoadClassByAlias('Customer');
    my $hFormData = $Servlet->form->formData($CustomerClass, 'UnitySaveBillingAddress');
    my $FormFields;
    foreach my $FormField (keys %{$hFormData->{'Fields'}}) {
        my $NewFormField = $FormField;
        $NewFormField =~ s/\./\_/;
        $FormFields->{$NewFormField} = $hFormData->{'Fields'}->{$FormField};
    }
    $Servlet->vars('FormFields', $FormFields);

    # check if we have both zipcode and city visible
    my $BillingAddressSetVisibleFields = $Shop->get('BillingAddressSetVisibleFields');
    my @BillingAddressSetVisibleFieldsArray = split /,/, $BillingAddressSetVisibleFields;
    my $ZipCodeAndCity = 'Zipcode' ~~ @BillingAddressSetVisibleFieldsArray && 'City' ~~ @BillingAddressSetVisibleFieldsArray;
    $Servlet->vars('ZipCodeAndCity', $ZipCodeAndCity);

}

sub UnityNewCustomer {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object;

    # create Customer and change object to customer
    # for saving the customer details (alias, address ect.) by change action UnitySaveCustomer
    my $Customer;
    Transaction( sub {
        # insert customer
        my $CustomerFolder = GetCustomerFolder($Shop);
        my $CustomerClass = $Shop->get('CustomerClass');

        my $NewAlias = $Servlet->form->value('Alias');

        if($NewAlias eq GetNextNo($Shop, 'Customer', $CustomerFolder, 0) ){
            $NewAlias = GetNextNo($Shop, 'Customer');
        }

        $Customer = $self->NewObject( $Servlet, $CustomerFolder, $CustomerClass, { 'Alias' => $NewAlias } );

        # insert address
        my $BillingAddress = $self->NewObject( $Servlet, $Customer, LoadClassByAlias('Address') );
        $Customer->set({'BillingAddress' => $BillingAddress});
    });

    # set customer to default next view object
    $Servlet->vars('ObjectID', $Customer->id);
    $Servlet->vars('ChangeObjectID', $Customer->id);
    $Servlet->vars('ViewObjectID', $Customer->id);
    $Servlet->vars('IsNewCustomer', 0);

}

sub CheckRegistrationFields {
    my $self = shift;
    my ($Servlet) = @_;
    my $hFieldNames = REGISTRATION_FIELDS_MAX_LENGTH();

    my @aWhiteList = qw( URL EMail EMailPrivate EMailBusiness );

    foreach my $FieldName (keys %{$hFieldNames}) {
        my $MaxLength = $hFieldNames->{$FieldName};
        my $FieldValue = $Servlet->form->value($FieldName);

        next unless isValid($FieldValue);

        if( length $FieldValue > $MaxLength ) {
            $Servlet->form->executeFormError({
                'Reason'=> 'INPUT_FIELD_FIELD_TOO_LONG',
                'Name'  => $FieldName,
            }) ;
            return;
        }

        next if ElementOf($FieldName, @aWhiteList);

        my @SpamFieldValues = split(/,/, SPAM_FIELD_VALUES);

        if ($FieldValue =~ /http/i or $FieldValue =~ /www/i or $FieldValue =~ /\/\// or $FieldValue =~ /@/) {
            $Servlet->form->executeFormError({
                'Reason'=> 'INVALID_REGISTRATION_INPUT_FIELD_FORMAT',
                'Name'  => $FieldName,
            }) ;
            return;
        }
        if ( scalar @SpamFieldValues ) {
            foreach my $SpamValue (@SpamFieldValues) {
                if ($FieldValue =~ qr/$SpamValue/i) {
                    $Servlet->form->executeFormError({
                        'Reason'=> 'INVALID_REGISTRATION_INPUT_FIELD_FORMAT',
                        'Name'  => $FieldName,
                    }) ;
                    return;
                }
            }
        }
    }
}

sub UnityViewCustomerGroups {
    my $self = shift;
    my ($Servlet) = @_;

    # enable language switch
    $Servlet->vars('ShowLanguageSwitch', 1) unless $Servlet->params('DELETE');

    $self->SUPER::View($Servlet);
}

1;
