#========================================================================================
# package      DE_EPAGES::Installer::API::LangDiffBuilder
# state        public
# description  This module generates and changes all files necessary for language patches
#
#========================================================================================
package DE_EPAGES::Installer::API::LangDiffBuilder;
use base qw( Exporter );

use strict;

use File::Basename;
use Getopt::Long;
use Data::Dumper;
use Config::IniFiles;
use DE_EPAGES::Core::API::CVS;
use DE_EPAGES::Core::API::File qw (
    GetFileNames
    ExistsDirectory
    ExistsFile
    CreateDirectory
    DeleteDirectory
    RunOnDirectory
    GetFileMD5Digest
    CopyFile
    DeleteFile
    OpenFile
    GetFileContent
);
use DE_EPAGES::Core::API::TestParameter qw( TestString );
use DE_EPAGES::Installer::API::XMLFilePatcher;
use DE_EPAGES::Installer::API::PatchXMLExportDriver;
use DE_EPAGES::Object::API::XMLCompare qw (CompareXML);
use DE_EPAGES::Object::API::XMLStructureExportDriver;

our @EXPORT_OK = qw (
    PostProcessXMLFile
    AddFilePrefix
    ChangeStructLanguage
    GetMainLanguage
    ReplaceLanguageAttribute
    ReplaceLanguageFileTag
    IsLanguageSupported
);

## no critic qw(RequireExtendedFormatting ProhibitLongLines)

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LangDiffBuilder = DE_EPAGES::Installer::API::LangDiffBuilder->new(
#                   'OldReleaseTag' => $OldReleaseTag,
#                   'NewReleaseTag' => $NewReleaseTag,
#                   'Lang' => $Lang,
#                   'Target' => $Target,
#                   'Verbose' => 1,
#               );
#----------------------------------------------------------------------------------------
# description  Creates a new LangDiffBuilder object
#----------------------------------------------------------------------------------------
# input        $options | keys
#                   <ul>
#                       <li>OldReleaseTag - old release tag, e.g. 'release_6_9_1'
#                       <li>NewReleaseTag - new release tag, e.g. 'release_6_9_2'
#                       <li>Lang          - language to create a diff for, e.g. 'de'
#                       <li>Target        - target directory for output diff files
#                       <li>Verbose       - print debug information (0 or 1)
#                   </ul>
#               | hash
# return       $LangDiffBuilder | LangDiffBuilder object | object
#========================================================================================
sub new {
    my $class = shift;
    my %options = @_;
    TestString( 'OldReleaseTag', $options{'OldReleaseTag'} );
    TestString( 'NewReleaseTag', $options{'NewReleaseTag'} );
    my $self = bless \%options, $class;

    $self->{'CompanyName'} = 'DE_EPAGES' if (not defined($self->{'CompanyName'}));
    $self->{'PatchStructure'} = {};

    return $self;
}

#========================================================================================
# function     exportReleaseFolders
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportReleaseFolders($CVSPath, $OldDir, $NewDir);
#----------------------------------------------------------------------------------------
# description  step 1
#               export two versions of one cartridge, one tagged with old release and
#               one tagged with the new release tag to different folders
#----------------------------------------------------------------------------------------
# input        $CVSPath | cvs path from cartridge that should be patcht | string
# input        $OldDir | directory to export cartridge version tagged with old release tag | string
# input        $NewDir | directory to export cartridge version tagged with new release tag | string
#========================================================================================
sub exportReleaseFolders {
    my $self = shift;
    my ($CVSPath, $OldDir, $NewDir) = @_;

    #print message
    $self->printMessage( "\n >step-1: export old and new version of cartridge to\n   -$OldDir and\n   -$NewDir\n" );

    #export cartridge versions tagged with old and new release
    $self->exportReleaseFolder($CVSPath, $OldDir, $self->{'OldReleaseTag'});
    $self->exportReleaseFolder($CVSPath, $NewDir, $self->{'NewReleaseTag'});
    return;
}

#========================================================================================
# function     generatePatchFiles
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->generatePatchFiles($OldFolder, $NewFolder, $CartridgeFolder);
#----------------------------------------------------------------------------------------
# description  step 2
#               - compare old and new version of cartridge
#               - generates all patch files under ../Patches/Patch../Database/XML
#               - (if patch files already builded, only additional tags where added
#                  manually changes with stay)
#----------------------------------------------------------------------------------------
# input        $OldFolder | folder of cartridgeversion tagged with old release tag | string
# input        $NewFolder | folder of cartridgeversion tagged with new release tag | string
# input        $CartridgeFolder | folder of cartridge, where should commit | string
#========================================================================================
sub generatePatchFiles {
    my $self = shift;
    my ($OldFolder, $NewFolder, $CartridgeFolder) = @_;

    #get name of folder to save patch files
    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->target);

    #print Message
    $self->printMessage("\n >step 2: collect information from old and new version of cartridge, save patch files at\n   $TargetFolder\n");

    #handle files that changed from old to new release tag
    $self->_handleChangedFiles($CartridgeFolder, $OldFolder, $NewFolder);

    #handle files that only can find in new cartridge version
    $self->_handleAddedFiles($CartridgeFolder, $OldFolder, $NewFolder);

    #handle files that only can find in old cartridge version
    $self->_handleDeletedFiles($CartridgeFolder, $OldFolder, $NewFolder);
}

#========================================================================================
# function     exportReleaseFolder
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportReleaseFolder($CVSPath, $Dir, $Tag);
#----------------------------------------------------------------------------------------
# description  part of step 1
#               export two versions of one cartridge, one tagged with old release and
#               one tagged with the new release tag to different folders
#----------------------------------------------------------------------------------------
# input        $CVSPath | cvs path from cartridge that should be patcht | string
# input        $OldDir | directory to export cartridge version tagged with $Tag | string
# input        $Tag | tag of cartridge, most any release tag | string
#========================================================================================
sub exportReleaseFolder {
    my $self = shift;
    my ($CVSPath, $Dir, $Tag) = @_;

    DeleteDirectory($Dir, { 'recursive'=>1});
    CreateDirectory($Dir);
    my $CVS = DE_EPAGES::Core::API::CVS->new({ '-q' => '' }, {'print' => 1});

    # Fix for Linux error "failed with exit code 256: cvsnt [export aborted]: absolute pathname [...] illegal for server"
    # run the command in the parent directory
    my $TopLevelDirectory = $Dir;
    $TopLevelDirectory =~ s/(?:.*)\/(.*)$/$1/;

    RunOnDirectory("$Dir/..", sub { $CVS->command('export', {'-d' => "./$TopLevelDirectory", '-r'=>$Tag}, [ $CVSPath ]) });
    return;
}

#========================================================================================
# function     verbose
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->verbose()
#----------------------------------------------------------------------------------------
# description  return object attribute verbose
#----------------------------------------------------------------------------------------
# return       $Verbose | object attribute | boolean
#========================================================================================
sub verbose {
    return shift->{'Verbose'};
}

#========================================================================================
# function     language
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->language()
#----------------------------------------------------------------------------------------
# description  return object attribute language
#----------------------------------------------------------------------------------------
# return       $Language | object attribute | string
#========================================================================================
sub language {
    return shift->{'Language'};
}

#========================================================================================
# function     target
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->target()
#----------------------------------------------------------------------------------------
# description  return object attribute target
#----------------------------------------------------------------------------------------
# return       $Target | object attribute | string
#========================================================================================
sub target {
    return shift->{'Target'};
}

#========================================================================================
# function     printMessage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->printMessage('message to test');
#----------------------------------------------------------------------------------------
# description  print message if verbose is true
#----------------------------------------------------------------------------------------
# input        $Message | message to print | string
#========================================================================================
sub printMessage {
    my $self = shift;
    my ($Message) = @_;

    if( $self->verbose ) {
        print $Message, "\n";
    }
}

#========================================================================================
# function     newVersion
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $NewVersion = $self->newVersion();
#----------------------------------------------------------------------------------------
# description  return object attribute, new version
#----------------------------------------------------------------------------------------
# return       $NewVersion | object attribute | string
#========================================================================================
sub newVersion {
    my $self = shift;
    return _GetVersionByReleaseTag( $self->{'NewReleaseTag'} );
}

#========================================================================================
# function     oldVersion
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $OldVersion = $self->oldVersion();
#----------------------------------------------------------------------------------------
# description  return object attribute, old version
#----------------------------------------------------------------------------------------
# return       $OldVersion | object attribute | string
#========================================================================================
sub oldVersion {
    my $self = shift;
    return _GetVersionByReleaseTag( $self->{'OldReleaseTag'} );
}

#========================================================================================
# function     _handleChangedFiles
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_handleChangedFiles($CartridgeFolder, $OldFolder, $NewFolder);
#----------------------------------------------------------------------------------------
# description  part of step 2
#               handle changed files, generates patch xml files
#----------------------------------------------------------------------------------------
# input        $CartridgeFolder | folder of cartridge, where should commit | string
# input        $OldFolder | folder of cartridgeversion tagged with old release tag | string
# input        $NewFolder | folder of cartridgeversion tagged with new release tag | string
#========================================================================================
sub _handleChangedFiles {
    my $self = shift;
    my ($CartridgeFolder, $OldFolder, $NewFolder) = @_;

    #get old files, by run on old directory
    my $aOldFiles = ExistsDirectory( $OldFolder )
        ? RunOnDirectory( $OldFolder, sub { GetFileNames( undef , {'recursive' => 1} ) } )
        : [];

    #get new files, by run on new directory
    my $aNewFiles = RunOnDirectory( $NewFolder, sub { GetFileNames( undef, {'recursive' => 1} ); } );

    #get all filenames that can find in the new and in the old directory
    my $aIntersectionItems = _GetIntersectionItems($aOldFiles,$aNewFiles);

    #remove files that are equal in old and new folder
    my $aChangedFiles = _RemoveFilesNotChanged($aIntersectionItems, $OldFolder, $NewFolder);

    #get cartridge name
    my $CartridgeName = $CartridgeFolder;
    my $CompanyName = $self->{'CompanyName'}; ## no critic qw( ProhibitUnusedVariablesEnhanced )
    $CartridgeName =~ s/^.*[\/\\]Cartridges[\/\\]$CompanyName[\/\\]//;

    #get patch target folder
    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->target);

    my $Language = $self->language;
    foreach my $FileName (@$aChangedFiles) {#for each changed file
        if(($FileName =~ m/.\.*$Language\.xml$/) and ($FileName !~ m/Patches\/.*/)) {
                my $OldFile = $OldFolder . '/' . $FileName;
                my $NewFile = $NewFolder . '/' . $FileName;
                $self->_handleChangedXMLFile( $CartridgeName, $OldFile, $NewFile, $FileName, $TargetFolder );
        }
    }
}

sub _handleChangedXMLFile {
    my $self = shift;
    my ($CartridgeName, $OldFile, $NewFile, $FileName, $TargetFolder) = @_;

    return if ($FileName =~ /\/SQL\/.*\.xml/);
    return if ($FileName =~ /\/mysql\/.*\.xml/);
    return if ($FileName =~ /^t\/.*\.xml/);
    return if ($FileName =~ /\/\.\#.*/);

    my $FilePatcher = DE_EPAGES::Installer::API::XMLFilePatcher->new();

    my $DirName = dirname($FileName);
    my $PatchFile = basename($FileName);

    my $DelPatchFile = $PatchFile;
    my $Language = $self->language;
    $DelPatchFile =~ s/(.*)\.$Language\.xml$/$1.xx.xml/g;

    #make delete file
    my $OrigPatchFile = "$TargetFolder/$DirName/transdel_$DelPatchFile";
    my $TmpPatchFile = "$TargetFolder/$DirName/transdel_$DelPatchFile";
    my $OldPatchFile = "$TargetFolder/$DirName/transdel_$DelPatchFile";
    $TmpPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(tmp\)$1/;
    $OldPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(old\)$1/;

    $self->{'PatchStructure'}->{'ChangeLog'}->{'Changed'}->{$CartridgeName . '/' . $FileName} = 1;

    #check if patch is already build
    if (not ExistsFile($OrigPatchFile)){
        #patch is build first time, build patch file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $OrigPatchFile, 1, {'register_notrans' => 1, 'quotecharacter' => '"'});

        #Cartridge tag have to add to shopchild dependency file
        $self->_addCartridgeTag($OrigPatchFile,$self->{'CompanyName'} . '::' . $CartridgeName) if (
            ExistsFile($OrigPatchFile) and
            $OrigPatchFile =~ /ShopChildDependencies\.xml$/
        );
    }
    else{
        #patch is already build, save new patch file as temporary file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $TmpPatchFile, 1, {'register_notrans' => 1, 'quotecharacter' => '"'});

        if (not ExistsFile($TmpPatchFile)){
            #new patch have no difference in file, print message
            print "\n    Warning: new version of patch dont have $OrigPatchFile included\n" ;
        }
        else{#a new and a old patch file exists look for witch is the right one
            #Cartridge tag have to add to shopchild dependency file
            $self->_addCartridgeTag($TmpPatchFile,$self->{'CompanyName'} . '::' . $CartridgeName)
                if ($TmpPatchFile =~ /ShopChildDependencies\.xml$/);

            if(CompareXML($TmpPatchFile, $OrigPatchFile, {'checkreference' => '1', 'checkdelete' => '1'})){
                #no changes have to add
                DeleteFile($TmpPatchFile);
            }
            else{
                my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
                my $OrigPatchFileXMLStruct = $ImportHandler->getXMLStructure($OrigPatchFile);
                my $TmpPatchFileXMLStruct = $ImportHandler->getXMLStructure($TmpPatchFile);

                $OrigPatchFileXMLStruct->assimilateXMLStructure($TmpPatchFileXMLStruct,{'overwrite' => '1'});
                my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
                CopyFile($OrigPatchFile,$OldPatchFile);
                $XMLStructExporter->exportXML($OrigPatchFile,$OrigPatchFileXMLStruct);
                DeleteFile($TmpPatchFile);

                print "\n    [Warning:] patch already build, please check changes\n      -previous $OldPatchFile\n      -new $OrigPatchFile\n";
            }
        }
    }

    #make import file
    $OrigPatchFile = $TargetFolder . '/' . $DirName . '/transimp_' . $PatchFile;
    $TmpPatchFile = $TargetFolder . '/' . $DirName . '/transimp_' . $PatchFile;
    $OldPatchFile = $TargetFolder . '/' . $DirName . '/transimp_' . $PatchFile;
    $TmpPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(tmp\)$1/;
    $OldPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(old\)$1/;

    #check if patch is already build
    if (not ExistsFile($OrigPatchFile)){
        #patch is build first time, build patch file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $OrigPatchFile, 0, {'register_notrans' => 1, 'quotecharacter' => '"'});
        #CartridgeTag have to add to shopchild dependency file
        $self->_addCartridgeTag($OrigPatchFile,$self->{'CompanyName'} . '::' . $CartridgeName) if (
            ExistsFile($OrigPatchFile) and
            $OrigPatchFile =~ /ShopChildDependencies\.xml$/
        );
    }
    else{
        #patch is already build, save new patch file as temporary file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $TmpPatchFile, 0, {'register_notrans' => 1, 'quotecharacter' => '"'});
        #CartridgeTag have to add to shopchild dependency file
        $self->_addCartridgeTag($TmpPatchFile,$self->{'CompanyName'} . '::' . $CartridgeName)
            if ($TmpPatchFile =~ /ShopChildDependencies\.xml$/);
        if (not ExistsFile($TmpPatchFile)){
            #new patch have no difference in file, print message
            print "\n    [Warning:] new version of patch dont have $OrigPatchFile included\n" ;
        }
        else{#a new and a old patch file exists look for witch is the right one
            if(CompareXML($TmpPatchFile, $OrigPatchFile, {'checkreference' => '1', 'checkdelete' => '1'})){
                DeleteFile($TmpPatchFile);
            }
            else{
                my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
                my $OrigPatchFileXMLStruct = $ImportHandler->getXMLStructure($OrigPatchFile);
                my $TmpPatchFileXMLStruct = $ImportHandler->getXMLStructure($TmpPatchFile);

                $OrigPatchFileXMLStruct->assimilateXMLStructure($TmpPatchFileXMLStruct,{'overwrite' => '1'});
                my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
                CopyFile($OrigPatchFile,$OldPatchFile);
                $XMLStructExporter->exportXML($OrigPatchFile,$OrigPatchFileXMLStruct);
                DeleteFile($TmpPatchFile);

                print "\n    [Warning:] patch already build, please check changes\n      -new $OrigPatchFile\n      -previous $OldPatchFile\n";
            }
        }
    }
}

#funktion handhabt alle hinzugefgten Files
sub _handleAddedFiles{
    my $self = shift;
    my ($CartridgeFolder, $OldFolder, $NewFolder) = @_;

    my $aOldFiles = ExistsDirectory( $OldFolder )
        ? RunOnDirectory( $OldFolder, sub { GetFileNames( undef , {'recursive' => 1} ) } )
        : [];

    my $aNewFiles = RunOnDirectory( $NewFolder, sub { GetFileNames( undef, {'recursive' => 1} ); } );

    my $aImportFiles = _GetMissingItems($aOldFiles,$aNewFiles);

    my $CartridgeName = $CartridgeFolder;
    my $CompanyName = $self->{'CompanyName'}; ## no critic qw( ProhibitUnusedVariablesEnhanced )
    $CartridgeName =~ s/^.*[\/\\]Cartridges[\/\\]$CompanyName[\/\\]//;

    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->target);

    my $Language = $self->language;
    foreach my $FileName (@$aImportFiles) {
        if(($FileName =~ m/.*\.$Language\.xml$/) and ($FileName !~ m/Patches\/.*/)) {
            my $NewFile = $NewFolder . '/' . $FileName;
            $self->_handleNewXMLFile( $CartridgeName, $NewFile, $FileName, $TargetFolder );
        }
    }
}

sub _handleNewXMLFile {
    my $self = shift;
    my ($CartridgeName, $NewFile, $FileName, $TargetFolder) = @_;

    return if ($FileName =~ /\/SQL\/.*\.xml/);
    return if ($FileName =~ /\/mysql\/.*\.xml/);
    return if ($FileName =~ /^t\/.*\.xml/);
    return if ($FileName =~ /\/\.\#.*/);

    my $FilePatcher = DE_EPAGES::Installer::API::XMLFilePatcher->new();

    my $OldFile = $ENV{'EPAGES_CARTRIDGES'} . '/DE_EPAGES/Installer/Scripts/empty.xml';

    my $DirName = dirname($FileName);
    my $PatchFile = basename($FileName);

    #make import file
    my $OrigPatchFile = "$TargetFolder/$DirName/transimp_$PatchFile";
    my $TmpPatchFile = "$TargetFolder/$DirName/transimp_$PatchFile";
    my $OldPatchFile = "$TargetFolder/$DirName/transimp_$PatchFile";
    $TmpPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(tmp\)$1/;
    $OldPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(old\)$1/;

    $self->{'PatchStructure'}->{'ChangeLog'}->{'New'}->{$CartridgeName . '/' . $FileName} = 1;

    #check if patch is already build
    if (not ExistsFile($OrigPatchFile)){
        #patch is build first time, build patch file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $OrigPatchFile, 0, {'register_notrans' => 1, 'quotecharacter' => '"' } );
        #CartridgeTag have to add to shopchild dependency file
        $self->_addCartridgeTag($OrigPatchFile,$self->{'CompanyName'} . '::' . $CartridgeName) if (
            ExistsFile($OrigPatchFile) and
            $OrigPatchFile =~ /ShopChildDependencies\.xml$/
        );
    }
    else{
        #patch is already build, save new patch file as temporary file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $TmpPatchFile, 0, {'register_notrans' => 1, 'quotecharacter' => '"' } );

        if (not ExistsFile($TmpPatchFile)){
            #new patch have no difference in file, print message
            print "\n    Warning: new version of patch dont have $OrigPatchFile included\n" ;
        }
        else{#a new and a old patch file exists look for witch is the right one
            #cartridge tag have to add to shopchild dependency file
            $self->_addCartridgeTag($TmpPatchFile,$self->{'CompanyName'} . '::' . $CartridgeName)
                if ($TmpPatchFile =~ /ShopChildDependencies\.xml$/);

            if(CompareXML($TmpPatchFile, $OrigPatchFile, {'checkreference' => '1', 'checkdelete' => '1'})){
                DeleteFile($TmpPatchFile);
            }
            else{
                my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
                my $OrigPatchFileXMLStruct = $ImportHandler->getXMLStructure($OrigPatchFile);
                my $TmpPatchFileXMLStruct = $ImportHandler->getXMLStructure($TmpPatchFile);

                $OrigPatchFileXMLStruct->assimilateXMLStructure($TmpPatchFileXMLStruct,{'overwrite' => '1'});
                my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
                CopyFile($OrigPatchFile,$OldPatchFile);
                $XMLStructExporter->exportXML($OrigPatchFile,$OrigPatchFileXMLStruct);
                DeleteFile($TmpPatchFile);

                print "\n    [Warning:] patch already build, please check changes\n      -previous $OrigPatchFile\n      -new $TmpPatchFile\n";
            }
        }
    }
}

#funktion handhabt alle gelschten files
sub _handleDeletedFiles{
    my $self = shift;
    my ($CartridgeFolder, $OldFolder, $NewFolder) = @_;

    my $aOldFiles = ExistsDirectory( $OldFolder )
        ? RunOnDirectory( $OldFolder, sub { GetFileNames( undef , {'recursive' => 1} ) } )
        : [];

    my $aNewFiles = RunOnDirectory( $NewFolder, sub { GetFileNames( undef, {'recursive' => 1} ); } );

    my $aDeleteFiles = _GetMissingItems($aNewFiles,$aOldFiles);

    my $CartridgeName = $CartridgeFolder;
    my $CompanyName = $self->{'CompanyName'}; ## no critic qw( ProhibitUnusedVariablesEnhanced )
    $CartridgeName =~ s/^.*[\/\\]Cartridges[\/\\]$CompanyName[\/\\]//;

    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->target);

    my $Language = $self->language;
    foreach my $FileName (@$aDeleteFiles) {
        if(($FileName =~ m/.*\.$Language\.xml$/) and ($FileName !~ m/^Patches\/.*/)) {
            my $OldFile = $OldFolder . '/' . $FileName;
            $self->_handleDeletedXMLFile( $CartridgeName, $OldFile, $FileName, $TargetFolder );
        }
    }
}

sub _handleDeletedXMLFile {
    my $self = shift;
    my ($CartridgeName, $OldFile, $FileName, $TargetFolder) = @_;

    my $NewFile = $ENV{'EPAGES_CARTRIDGES'} . '/DE_EPAGES/Installer/Scripts/empty.xml';

    my $FilePatcher = DE_EPAGES::Installer::API::XMLFilePatcher->new();

    my $DirName = dirname($FileName);
    my $PatchFile = basename($FileName);

    my $DelPatchFile = $PatchFile;
    my $Language = $self->language;
    $DelPatchFile =~ s/(.*)\.$Language\.xml$/$1.xx.xml/g;

    #make delete file
    my $OrigPatchFile = "$TargetFolder/$DirName/transdelfile_$DelPatchFile";
    my $TmpPatchFile = "$TargetFolder/$DirName/transdelfile_$DelPatchFile";
    my $OldPatchFile = "$TargetFolder/$DirName/transdelfile_$DelPatchFile";
    $TmpPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(tmp\)$1/;
    $OldPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(old\)$1/;

    $self->{'PatchStructure'}->{'ChangeLog'}->{'Deleted'}->{$CartridgeName . '/' . $FileName} = 1;

    #check if patch is already build
    if (not ExistsFile($OrigPatchFile)){
        #patch is build first time, build patch file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $OrigPatchFile, 1, {'register_notrans' => 1, 'quotecharacter' => '"'});
        #CartridgeTag have to add to shopchild dependency file
        $self->_addCartridgeTag($OrigPatchFile,$self->{'CompanyName'} . '::' . $CartridgeName) if (
            ExistsFile($OrigPatchFile) and
            $OrigPatchFile =~ /ShopChildDependencies\.xml$/
        );
    }
    else{
        #patch is already build, save new patch file as temporary file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $TmpPatchFile, 1, {'register_notrans' => 1, 'quotecharacter' => '"'});
        if (not ExistsFile($TmpPatchFile)){
            #new patch have no difference in file, print message
            print "\n    Warning: new version of patch dont have $OrigPatchFile included\n" ;
        }
        else{#a new and a old patch file exists look for witch is the right one
            #CartridgeTag have to add to shopchild dependency file
            $self->_addCartridgeTag($TmpPatchFile,$self->{'CompanyName'} . '::' . $CartridgeName)
                if ($TmpPatchFile =~ /ShopChildDependencies\.xml$/);

            if(CompareXML($TmpPatchFile, $OrigPatchFile, {'checkreference' => '1', 'checkdelete' => '1'})){
                DeleteFile($TmpPatchFile);
            }
            else{
                my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
                my $OrigPatchFileXMLStruct = $ImportHandler->getXMLStructure($OrigPatchFile);
                my $TmpPatchFileXMLStruct = $ImportHandler->getXMLStructure($TmpPatchFile);

                $OrigPatchFileXMLStruct->assimilateXMLStructure($TmpPatchFileXMLStruct,{'overwrite' => '1'});
                my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
                CopyFile($OrigPatchFile,$OldPatchFile);
                $XMLStructExporter->exportXML($OrigPatchFile,$OrigPatchFileXMLStruct);
                DeleteFile($TmpPatchFile);

                print "\n    [Conflict:] patch already build\n      -previous $OrigPatchFile\n      -new $TmpPatchFile\n";
            }
        }
    }
}

#HilfsFunktion ermittelt alle Namen der Ordner die fehlen
sub _GetMissingItems {
    my ($aReferenceNames,$aControlNames) = @_;

    my $hRef;
    my @Result = ();
    my $CleanName;

    foreach my $RefName (@$aReferenceNames){
        $CleanName = $RefName;
        $CleanName =~ s/^\.//;
        $hRef->{$CleanName} = '1';
    }

    foreach my $ControlName (@$aControlNames){
        $CleanName = $ControlName;
        $CleanName =~ s/^\.//;
        push(@Result, $CleanName) if ($hRef->{$CleanName} ne '1');
    }

    return \@Result;
}

#gibt die Namen zurck die in beiden Arrays enthalten sind
sub _GetIntersectionItems {
    my ($aReferenceNames,$aControlNames) = @_;

    my $hRef;
    my @Result = ();
    my $CleanName;

    foreach my $RefName (@$aReferenceNames){
        $CleanName = $RefName;
        $CleanName =~ s/^\.//;
        $hRef->{$CleanName} = '1';
    }

    foreach my $ControlName (@$aControlNames){
        $CleanName = $ControlName;
        $CleanName =~ s/^\.//;
        push(@Result, $CleanName) if ($hRef->{$CleanName} eq '1');
    }

    return \@Result;
}

sub _RemoveFilesNotChanged{
    my ($aIntersectionItems, $OldDir, $NewDir) = @_;

    my @Result;

    foreach my $FileName (@$aIntersectionItems){
        my $CheckSumOld = GetFileMD5Digest($OldDir . $FileName);
        my $CheckSumNew = GetFileMD5Digest($NewDir . $FileName);
        push(@Result, $FileName) if ($CheckSumOld ne $CheckSumNew);
    }
    return \@Result;
}

sub _GetVersionByReleaseTag{
    my ($ReleaseTag) = @_;
    my $CartridgeVersion = $ReleaseTag;
    $CartridgeVersion =~ s/^(?:lp_|cs_)?release_//;
    $CartridgeVersion =~ s/\_/\./g;

    return $CartridgeVersion;
}

sub _GetPatchTargetName {
    my ($CartridgeFolder, $TargetDir) = @_;

    my $CartridgeName = $CartridgeFolder;
    $CartridgeName =~ s/^.*[\/\\]Cartridges//;

    my $Target = $TargetDir;
    #$Target =~ s/^[\/\\]//;
    $Target =~ s/[\/\\]$//;

    return $Target . $CartridgeName;
}

sub _addCartridgeTag{
    my $self = shift;
    my ($PatchFile, $CartridgeName) = @_;

    my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
    my $XMLStruct = $ImportHandler->getXMLStructure($PatchFile);

    foreach my $Element (@{$XMLStruct->getElements()}){
        next if($XMLStruct->getXMLStructAttribute($Element,'Position') eq '1');
        $XMLStruct->setXMLStructAttribute($Element,'Position',$XMLStruct->getXMLStructAttribute($Element,'Position') + 1 );
    }

    my $CompanyName = $self->{'CompanyName'};

    $XMLStruct->addElement(
        '/Cartridge',
        {},
        {
            'Position' => '2',
            'TagName' => 'Cartridge',
            'Class' => 'Cartridge',
            'NonEffectiveForSubPathes' => '1'
        },
        {'Package' => "$CompanyName::" . $CartridgeName}
    );

    my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
    $XMLStructExporter->exportXML($PatchFile,$XMLStruct);
}

#========================================================================================
# function     PostProcessXMLFile
# state        public
#----------------------------------------------------------------------------------------
# syntax       PostProcessXMLFile($FileName);
#----------------------------------------------------------------------------------------
# description  unquotes &lt;notrans&gt;, &quot; and &apos;
#----------------------------------------------------------------------------------------
# input        $FileName | XML file name | string
#========================================================================================
sub PostProcessXMLFile {
    my ($XMLFile) = @_;

    my $PrepFileName = $XMLFile . '__prep__';

    OpenFile('<', $XMLFile, sub {
        my ($InFile) = @_;
        OpenFile('>', $PrepFileName, sub {
            my ($OutFile) = @_;
            while (my $Line = <$InFile>) {
                $Line =~ s/&lt;notrans&gt;/<notrans>/g;
                $Line =~ s/&lt;\/notrans&gt;/<\/notrans>/g;
                while ($Line =~ /(>.*)&quot;(.*<)/) {
                    $Line =~ s/(>.*)&quot;(.*<)/$1"$2/; # &quot; between tags
                }
                while ($Line =~ /(>.*)&apos;(.*<)/) {
                    $Line =~ s/(>.*)&apos;(.*<)/$1'$2/; # &apos; between tags
                }
                while ($Line =~ /(<.*)&apos;(.*>)/) {
                    $Line =~ s/(<.*)&apos;(.*>)/$1'$2/; # &apos; in attribute
                }
                print $OutFile $Line;
            }
        });
    });

    unlink($XMLFile);
    rename($PrepFileName, $XMLFile);
}

#========================================================================================
# function     AddFilePrefix
# state        public
#----------------------------------------------------------------------------------------
# syntax       $NewFilePath = AddFilePrefix($FilePath, $Prefix);
#----------------------------------------------------------------------------------------
# description  adds a prefix to a file name in a path: /tmp/filename -> /tmp/prefixfilename/
#----------------------------------------------------------------------------------------
# input        $FilePath | full path to file | string
# input        $Prefix | prefix to prepend | string
# return       $NewFilePath | file path with file name prefixed | string
#========================================================================================
sub AddFilePrefix {
    my ($FileName, $Prefix) = @_;
    my $File = basename($FileName);
    my $Path = dirname($FileName);

    return $Path . '/' . $Prefix . $File;
}

#========================================================================================
# function     ChangeStructLanguage
# state        public
#----------------------------------------------------------------------------------------
# syntax       ChangeStructLanguage($Structure, $Language);
#----------------------------------------------------------------------------------------
# description  changes the language of an xml structure
#----------------------------------------------------------------------------------------
# input        $Structure | xml structure to change | object
# input        $Language | set the structure's language to this language | string
#========================================================================================
sub ChangeStructLanguage {
    my ($Structure, $Language) = @_;

    foreach my $Element (@{$Structure->getElements()}) {
        $Structure->setTagAttribute($Element, 'Language', $Language) if($Structure->getTagAttribute($Element, 'Language'));
        $Structure->setIDValue($Element, 'Language', $Language) if($Structure->getIDValue($Element, 'Language'));

        my $NewElementName = $Element;
        if($NewElementName =~ s/^(.*Language\->).*(;.*)$/$1$Language$2/x && $NewElementName ne $Element) {
            $Structure->renameElement($Element, $NewElementName);
        }
    }

    return;
}

#========================================================================================
# function     GetMainLanguage
# state        public
#----------------------------------------------------------------------------------------
# syntax       GetMainLanguage($LanguageIdentifier);
# example      GetMainLanguage('en-US')
#----------------------------------------------------------------------------------------
# description  returns the main language identifier (en-US -> en)
#----------------------------------------------------------------------------------------
# input        $Language | language identifier | string
# return       $MainLanguage | language identifier without region identifier | string
#========================================================================================
sub GetMainLanguage {
    # xx-YY -> xx
    my ($Language) = @_;
    if($Language =~ /(\w+)-/) {
        return $1;
    }

    return $Language;
}

#========================================================================================
# function     ReplaceLanguageAttribute
# state        public
#----------------------------------------------------------------------------------------
# syntax       ReplaceLanguageAttribute($XML);
# example      ReplaceLanguageAttribute($XML, 'en')
#----------------------------------------------------------------------------------------
# description  Replaces the <Language Language='xx'> attribute value in a string containing
#               epages dictionary XML
#----------------------------------------------------------------------------------------
# input        $XML | string containing XML | string
# input        $Language | language identifier to set | string
# return       $XML | changed XML | string
#========================================================================================
sub ReplaceLanguageAttribute {
    my ($XML, $Language) = @_;

    $XML =~ s/
        (\<Language   # opening <Language
        [^>]*         # other attributes
        Language      # Language attribute name
        \s*           # optional whitespace
        =             # =
        \s*           # optional whitespace
        ['"])         # opening quote
        [^'"]+        # the language identifier (e. g. 'de')
        (['"]         # closing quote
        [^>]*         # other attributes
        >)            # closing >
        /$1\Q$Language\E$2/x # replace language identifier
    ;

    return $XML;
}

#========================================================================================
# function     ReplaceLanguageFileTag
# state        public
#----------------------------------------------------------------------------------------
# syntax       ReplaceLanguageFileTag($Filename, $Language);
# example      ReplaceLanguageAttribute('transimp_Dictionary.de.xml', 'en') -> transimp_Dictionary.en.xml
#----------------------------------------------------------------------------------------
# description  Replaces the language tag in a translation XML file name
#----------------------------------------------------------------------------------------
# input        $Filename | filename to operate on | string
# input        $Language | the language identifier to set | string
# return       $Filename | changed file name | string
#========================================================================================
sub ReplaceLanguageFileTag {
    my ($Filename, $Language) = @_;

    $Filename =~ s/(\w+\.)(\w+)(\.xml)$/$1$Language$3/;
    return $Filename;
}

#========================================================================================
# function     IsLanguageSupported
# state        public
#----------------------------------------------------------------------------------------
# syntax       IsLanguageSupported($CartridgePath, $Language);
# example      IsLanguageSupported('/srv/epages/eproot/Cartridges/DE_EPAGES/Amazon', 'en')
#----------------------------------------------------------------------------------------
# description  Checks if the supplied language is supported by the Cartridge in $CartridgePath
#----------------------------------------------------------------------------------------
# input        $CartridgePath | folder of the cartridge | string
# input        $Language | the language identifier to check | string
# return       $Supported | 1 if the language is supported, otherwise 0 | boolean
#========================================================================================
sub IsLanguageSupported {
    my ($CartridgePath, $Language) = @_;

    my $InfoFile = "$CartridgePath/SupportedLanguages.txt";
    return 1 unless(ExistsFile($InfoFile));

    my %Ini;
    if (tie %Ini, 'Config::IniFiles', (-file => $InfoFile)) {
        # is ini file
        return ($Ini{$Language}{'supported'} eq '1') ? 1 : 0;
    } else {
        # is csv file
        my $SupportedLanguages = ${GetFileContent($InfoFile)} or die $!;
        my @Languages = split(',', $SupportedLanguages); ## no critic(ProhibitStringySplit)
        $Languages[$_] =~ s/^\s+|\s+$//g for (0..scalar(@Languages)-1); # trim
        return ($Language ~~ @Languages) ? 1 : 0;
    }
}

1;
