#========================================================================================
# package      DE_EPAGES::Zettle::API::WebService::PusherService
# base         DE_EPAGES::Zettle::API::WebService::RestService
# state        private
#----------------------------------------------------------------------------------------
# description  Service wrapper around Zettle Pusher API to handle webhook-related methods
#               See https://github.com/iZettle/api-documentation/blob/master/pusher-api/api-reference.md
#               See https://pusher.izettle.com/swagger
#----------------------------------------------------------------------------------------
# notes        Pusher is an API that publishes information to the integrator's service. This information is data related to products, purchases, inventory etc.
#               You can subscribe to specific events in the Pusher service to receive incoming HTTP POST notification messages when events occur.
#               The Pusher service uses Webhooks. To create a webhook, users configure a webhook listener and subscribe it to events.
#========================================================================================
package DE_EPAGES::Zettle::API::WebService::PusherService;
use base DE_EPAGES::Zettle::API::WebService::RestService;

use strict;

use HTTP::Status qw( HTTP_NOT_FOUND );

use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Zettle::API::Tools qw ( GenerateUUID );
use DE_EPAGES::Zettle::API::WebService::DTO::SubscriptionResponseDTO;
##no critic qw(RequireCorrectImports)
use DE_EPAGES::Zettle::API::WebService::EndpointHandler qw (
    CreateSubcriptionApiUrl
    DeleteSubscriptionApiUrl
    GetAllSubscriptionsApiUrl
);
##use critic qw(RequireCorrectImports)

use constant TRANSPORT_NAME => 'WEBHOOK';

# List of supported events. See https://github.com/iZettle/api-documentation/blob/master/pusher-api/api-reference.md#supported-events
# See also Models > CreateSubscription > eventNames in https://pusher.izettle.com/swagger#/subscriptions/createSubscription
use constant SUPPORTED_EVENTS => [qw(
    PurchaseCreated
    OrganizationUpdated
    ProductCreated
    ProductUpdated
    ProductDeleted
    InventoryBalanceChanged
    InventoryTrackingStarted
    InventoryTrackingStopped
    ApplicationConnectionRemoved
    PersonalAssertionDeleted
)];

# Map name of event to required scope (Note: '*' matches any scope)
our %EVENT_SCOPES = (
    'PurchaseCreated'          => 'READ:PURCHASE',
    'OrganizationUpdated'      => 'READ:USERINFO',
    'ProductCreated'           => 'READ:PRODUCT',
    'ProductUpdated'           => 'READ:PRODUCT',
    'ProductDeleted'           => 'READ:PRODUCT',
    'InventoryBalanceChanged'  => 'READ:PRODUCT',
    'InventoryTrackingStarted' => 'READ:PRODUCT',
    'InventoryTrackingStopped' => 'READ:PRODUCT',
    'ApplicationConnectionRemoved' => '*',
    'PersonalAssertionDeleted' => '*'
);


#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Zettle::API::WebService::PusherService->new($Environment);
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# input        $Environment | a Zettle environment object | DE_EPAGES::Zettle::API::Object:ZettleEnvironment
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Zettle::API::WebService::PusherService
#========================================================================================
sub new {
    my $class = shift;
    my ($Environment) = @_;

    my $self = $class->SUPER::new($Environment);
    return $self;
}

#========================================================================================
# function     createSubscription
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ProductInventoryDTO = $self->createSubscription($aEventNames);
#----------------------------------------------------------------------------------------
# description  Creates a webhook subscription to the given events.
#               See: https://pusher.izettle.com/swagger#/subscriptions/createSubscription
#----------------------------------------------------------------------------------------
# input        $destination | Callback URL where webhook will publish event data | string
# input        $contactEmail | Email to send information if something fails | string
# input        $aEventNames | optional. List of name of events to subscribe to.
#                              If parameter is not present you get subscribed to all events.
#                              See DE_EPAGES::Zettle::API:WebService:PusherService::%SUPORTED_EVENTS | ref.array.string
#----------------------------------------------------------------------------------------
# return       $SubscriptionResponseDTO | a SubscriptionResponseDTO object | DE_EPAGES::Zettle::API::WebService::DTO::SubscriptionResponseDTO
#========================================================================================
sub createSubscription {
    my $self = shift;

    my ($destination, $contactEmail, $aEventNames) = @_;

    $aEventNames //= (); # Default: empty array which means subscribe to all events

    my $CreateSubscriptionUrl = CreateSubcriptionApiUrl();
    my $hHeadersParams = {
        'Content-type' => 'application/json'
    };
    my $hContent = $self->_buildCreateSubscriptionContents($destination, $contactEmail, $aEventNames);

    my $HttpResponse = $self->sendPostRequest($CreateSubscriptionUrl, $hHeadersParams, $hContent);
    my $hResponse = $self->decodeResponseContent($HttpResponse);

    my $SubscriptionResponseDTO = DE_EPAGES::Zettle::API::WebService::DTO::SubscriptionResponseDTO->parseResponse($hResponse);

    return $SubscriptionResponseDTO;
}

#========================================================================================
# function     getAllSubscriptions
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aSubscriptionResponseDTO = $self->getAllSubscriptions();
#               $aSubscriptionResponseDTO = $self->getAllSubscriptions($ShopDomain);
#----------------------------------------------------------------------------------------
# description  Gets all the webhook subscriptions.
#               See: https://pusher.izettle.com/swagger#/subscriptions/getSubscriptions
#               See: https://github.com/iZettle/api-documentation/blob/master/pusher-api/api-reference.md#get-all-subscriptions
#----------------------------------------------------------------------------------------
# input        $ShopDomain | Optional. Get only the subscriptions that contains this domain (i.e. "zettle-es.example.com") | string
#----------------------------------------------------------------------------------------
# return       $aSubscriptionResponseDTO | List of all subscriptions (each element is a SubscriptionResponseDTO object)
#                                           See DE_EPAGES::Zettle::API::WebService::DTO::SubscriptionResponseDTO | ref.array.object
#========================================================================================
sub getAllSubscriptions {
    my $self = shift;
    
    my ($ShopDomain) = @_;

    my $GetAllSubscriptionsUrl = GetAllSubscriptionsApiUrl();
    my $HttpResponse = $self->sendGetRequest($GetAllSubscriptionsUrl);
    my $ahSubscriptionResponse = $self->decodeResponseContent($HttpResponse);
    my @aSubscriptionResponseDTO = ();
    foreach my $hSubscriptionResponse (@{$ahSubscriptionResponse}) {
        my $SubscriptionResponseDTO = DE_EPAGES::Zettle::API::WebService::DTO::SubscriptionResponseDTO->parseResponse($hSubscriptionResponse);
        if(!defined($ShopDomain) || $SubscriptionResponseDTO->destination =~ /$ShopDomain/){
            push(@aSubscriptionResponseDTO, $SubscriptionResponseDTO);
        }
    }

    return \@aSubscriptionResponseDTO;
}

#========================================================================================
# function     deleteSubscription
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Deleted = $self->deleteSubscription($SubscriptionUuid);
#----------------------------------------------------------------------------------------
# description  Deletes an existing webhook subscription.
#               See: https://pusher.izettle.com/swagger#/subscriptions/removeSubscription
#               See: https://github.com/iZettle/api-documentation/blob/master/pusher-api/api-reference.md#delete-a-subscription
#----------------------------------------------------------------------------------------
# input        $SubscriptionUuid | Unique identifier for an existing subscription (UUIDv1) | string
#----------------------------------------------------------------------------------------
# return       $Deleted | TRUE if subscription was deleted, FALSE if SubscriptionUuid was not found | boolean
#========================================================================================
sub deleteSubscription {
    my $self = shift;
    my ($SubscriptionUuid) = @_;

    my $DeleteSubscriptionUrl = DeleteSubscriptionApiUrl($SubscriptionUuid);

    eval {
        $self->sendDeleteRequest($DeleteSubscriptionUrl);
    };

    if(ExistsError()) {
        my $Error = GetError();
        if($Error->vars()->{'Status'} == HTTP_NOT_FOUND){
            return 0;
        }
        Error($Error); #forward all other Errors
    }

    return 1;
}


#========================================================================================

sub _buildCreateSubscriptionContents {
    my $self = shift;
    my ($destination, $contactEmail, $aEventNames) = @_;

    # Create a new unique identifier for subscription (UUIDv1)
    my $Uuid = GenerateUUID();

    return {
        'uuid' => $Uuid,
        'transportName' => TRANSPORT_NAME,
        'eventNames' => $aEventNames,
        'destination' => $destination,
        'contactEmail' => $contactEmail
    };
}

1;
