#========================================================================================
# package      DE_EPAGES::Zettle::API::WebService::RestService
# state        public
#----------------------------------------------------------------------------------------
# description  Generic REST Service for requesting to epagesj
#========================================================================================
package DE_EPAGES::Zettle::API::WebService::RestService;
use base Exporter;

use strict;

our @EXPORT_OK = qw ( WEB_SERVICE_ERROR_CODE );

use HTTP::Response;
use JSON qw ( decode_json encode_json );
use LWP::UserAgent;
use Time::HiRes qw ( usleep );
use URI;

use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Object::API::Factory qw ( LoadRootObject );
use DE_EPAGES::EPagesJ::API::Object::RestClient;
use DE_EPAGES::WebService::API::CodecRegistry;
use DE_EPAGES::Zettle::API::Logger;
use DE_EPAGES::Zettle::API::Tools qw ( GenerateCorrelationID );
use DE_EPAGES::Zettle::API::TokenHelper;
use DE_EPAGES::Zettle::API::WebService::AuthService;
use DE_EPAGES::Zettle::API::WebService::EndpointHandler qw ( GetAPICallRetryDelay );

# Define HTTP code 429 (missing in outdated HTTP::Status module)
use constant HTTP_TOO_MANY_REQUESTS => 429;
# Maximum number of retries for API calls before throwing an error
use constant MAX_RETRIES => 5;

# Http verbs
use constant HTTP_GET    => 'GET';
use constant HTTP_POST   => 'POST';
use constant HTTP_PUT    => 'PUT';
use constant HTTP_DELETE => 'DELETE';

# Error code for a web service response that resulted in failure
use constant REQUEST_FAILED_CODE => 'REQUEST_FAILED';
# Error code used to identify errors generated after a WebService operation
use constant WEB_SERVICE_ERROR_CODE => 'WebserviceError';

# For API token errors see https://github.com/iZettle/api-documentation/blob/master/authorization.md#errors
# Contents of the 'errorType' field for a Zettle HTTP response when AccessToken is expired
use constant ACCESS_TOKEN_EXPIRED_CODE => 'ACCESS_TOKEN_EXPIRED';
# Contents of the 'error' field for a Zettle HTTP response when RefreshToken is not valid
use constant INVALID_GRANT_CODE => 'invalid_grant';
# Contents of the 'error_description' field for a Zettle HTTP response when RefreshToken is not valid
use constant INVALID_REFRESH_TOKEN_CODE => 'invalid refresh_token';

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Zettle::API::WebService::RestService->new();
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# input        $Environment | a ZettleEnvironment object | DE_EPAGES::Zettle::API::Object:ZettleEnvironment
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Zettle::API::WebService::RestService
#========================================================================================
sub new {
    my $class = shift;
    my ($Environment) = @_;
    my $self = bless { 'Environment' => $Environment }, $class;

    return $self;
}

#========================================================================================
# function     getEnvironment
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->getEnvironment();
#----------------------------------------------------------------------------------------
# description  Retrieve ZettleEnvironment object.
#----------------------------------------------------------------------------------------
# return       $Environment | a ZettleEnvironment object | DE_EPAGES::Zettle::API::Object:ZettleEnvironment
#========================================================================================
sub getEnvironment {
    my $self = shift;
    return $self->{Environment};
}

#========================================================================================
# function     sendGetRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->sendGetRequest($Url, $hOptions);
#----------------------------------------------------------------------------------------
# description  sends an HTTP POST request
#----------------------------------------------------------------------------------------
# throws       JsonParserError, WebserviceError, HttpEncodingError
#----------------------------------------------------------------------------------------
# input        $Url | target endpoint URL | string
# input        $hOptions | optional. Additional request data with the following keys:
#                          <ul>
#                            <li>headers - optional. HTTP header data see <function DE_EPAGES::EPagesJ::API::Object::RestClient::new> - ref.hash
#                            <li>content - optional. content data - ref.hash
#                            <li>authenticated - optional. determines if request requires authorization header (default: true) - boolean
#                          </ul> | ref.hash
#----------------------------------------------------------------------------------------
# return       $hResponse | response as object | HTTP::Response
#========================================================================================
sub sendGetRequest {
    my $self = shift;
    my ($Url, $hOptions) = @_;

    $hOptions //= {};
    $hOptions->{'method'} = HTTP_GET;
    $hOptions->{'url'} = $Url;
    my $Response = $self->_sendRequestWithRefreshWrapper(%{$hOptions});

    return $Response;
}

#========================================================================================
# function     sendPostRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->sendPostRequest($Url, $hHeaders, $hContent);
#----------------------------------------------------------------------------------------
# description  sends an HTTP POST request
#----------------------------------------------------------------------------------------
# throws       JsonParserError, WebserviceError, HttpEncodingError
#----------------------------------------------------------------------------------------
# input        $Url | target endpoint URL | string
# input        $hHeaders | HTTP header data see <function DE_EPAGES::EPagesJ::API::Object::RestClient::new> | ref.hash
# input        $hContent | content data | ref.hash or arrayref
# input        $Authenticated | optional. determines if request requires authorization header (default: true) | boolean
#----------------------------------------------------------------------------------------
# return       $hResponse | response as object | HTTP::Response
#========================================================================================
sub sendPostRequest {
    my $self = shift;
    my ($Url, $hHeaders, $hContent, $Authenticated) = @_;
    
    my %hOptions = (
        'method'  => HTTP_POST,
        'url'     => $Url,
        'headers' => $hHeaders,
        'content' => $hContent,
    );
    $hOptions{'authenticated'} = $Authenticated if defined $Authenticated;
    my $Response = $self->_sendRequestWithRefreshWrapper(%hOptions);

    return $Response;
}

#========================================================================================
# function     sendPutRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->sendPutRequest($Url, $hHeaders, $hContent);
#----------------------------------------------------------------------------------------
# description  sends an HTTP PUT request
#----------------------------------------------------------------------------------------
# throws       JsonParserError, WebserviceError, HttpEncodingError
#----------------------------------------------------------------------------------------
# input        $Url | target endpoint URL | string
# input        $hHeaders | HTTP header data see <function DE_EPAGES::EPagesJ::API::Object::RestClient::new> | ref.hash
# input        $hContent | content data | ref.hash or arrayref
# input        $Authenticated | optional. determines if request requires authorization header (default: true) | boolean
#----------------------------------------------------------------------------------------
# return       $hResponse | response as object | HTTP::Response
#========================================================================================
sub sendPutRequest {
    my $self = shift;
    my ($Url, $hHeaders, $hContent, $Authenticated) = @_;
    
    my %hOptions = (
        'method'  => HTTP_PUT,
        'url'     => $Url,
        'headers' => $hHeaders,
        'content' => $hContent,
    );
    $hOptions{'authenticated'} = $Authenticated if defined $Authenticated;
    my $Response = $self->_sendRequestWithRefreshWrapper(%hOptions);

    return $Response;
}

#========================================================================================
# function     sendDeleteRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hResponse = $self->sendDeleteRequest($Url, $hOptions)
#----------------------------------------------------------------------------------------
# description  sends an HTTP DELETE request
#----------------------------------------------------------------------------------------
# throws       WebserviceError
#----------------------------------------------------------------------------------------
# input        $Url | target endpoint URL | string
# input        $hOptions | optional. Additional request data with the following keys:
#                          <ul>
#                            <li>headers - optional. HTTP header data see <function DE_EPAGES::EPagesJ::API::Object::RestClient::new> - ref.hash
#                            <li>content - optional. content data - ref.hash
#                            <li>authenticated - optional. determines if request requires authorization header (default: true) - boolean
#                          </ul> | ref.hash
#----------------------------------------------------------------------------------------
# return       $hResponse | response as object | HTTP::Response
#========================================================================================
sub sendDeleteRequest {
    my $self = shift;
    my ($Url, $hOptions) = @_;
    
    $hOptions //= {};
    $hOptions->{'method'} = HTTP_DELETE;
    $hOptions->{'url'} = $Url;
    my $Response = $self->_sendRequestWithRefreshWrapper(%{$hOptions});

    return $Response;
}


#========================================================================================
# function     decodeResponseContent
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->decodeResponseContent($HttpResponse);
#----------------------------------------------------------------------------------------
# description  Decode response content based on Content-Type.
#               Use <DE_EPAGES::WebService::API::CodecRegistry> to retrieve the appropiate <DE_EPAGES::WebService::API::Codec> object to decode HTTP content.
#----------------------------------------------------------------------------------------
# throws       HttpDecodingError
#----------------------------------------------------------------------------------------
# input        $HttpResponse | HTTP response object | HTTP::Response
#----------------------------------------------------------------------------------------
# return       $DecodedContent | decoded content | any
#========================================================================================
sub decodeResponseContent {
    my $self = shift;
    my ($HttpResponse) = @_;

    my $ContentType = $HttpResponse->content_type();
    # Default content type
    $ContentType = 'application/json' if not isValid($ContentType);
    
    my $DecodedContent;
    if($ContentType eq 'application/json'){
        $DecodedContent = decode_json($HttpResponse->content);
    } else {
        my $Codec = DE_EPAGES::WebService::API::CodecRegistry->get($ContentType);
        if($Codec){
            $DecodedContent = $Codec->decode($HttpResponse);
        } else {
            Error('HttpDecodingError', { 'Message' => "No codec found for Content-Type: $ContentType" });
        }
    }
    
    return $DecodedContent;
}


#========================================================================================


#========================================================================================
# function     _encodeContent
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_encodeContent($Content, $ContentType);
#----------------------------------------------------------------------------------------
# description  Use <DE_EPAGES::WebService::API::CodecRegistry> to retrieve the appropiate <DE_EPAGES::WebService::API::Codec> object to encode HTTP content.
#----------------------------------------------------------------------------------------
# throws       HttpEncodingError
#----------------------------------------------------------------------------------------
# input        $Content | HTTP raw body | any
# input        $ContentType | Optional. HTTP content type (default: application/json) | String
#----------------------------------------------------------------------------------------
# return       $EncodedContent | encoded content | any
#========================================================================================
sub _encodeContent {
    my $self = shift;
    my ($Content, $ContentType) = @_;

    # Default content type
    $ContentType = 'application/json' if not isValid($ContentType);
    
    my $Codec = DE_EPAGES::WebService::API::CodecRegistry->get($ContentType);
    if($Codec){
        my $Message = HTTP::Message->new();
        $Codec->encode($Message, $Content);
        return $Message->content();
    }else{
        Error('HttpEncodingError', { 'Message' => "No codec found for Content-Type: $ContentType" });
    }
    
    return;
}

#========================================================================================
# function     _sendRequestWithRefreshWrapper
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->_sendRequestWithRefreshWrapper($hOptions);
#----------------------------------------------------------------------------------------
# description  sends an HTTP request wrapping _sendRequest method so that AccessToken is
#               automatically refreshed if we get an access_token_expired error.
#----------------------------------------------------------------------------------------
# throws       JsonParserError, WebserviceError, HttpEncodingError, RetryLimitReached
#----------------------------------------------------------------------------------------
# input        $hOptions | arguments and options used to construct HTTP request with keys:
#                          <ul>
#                              <li>method - HTTP method: get/put/post/delete - string
#                              <li>url - target endpoint URL - string
#                              <li>headers - HTTP header data see <function DE_EPAGES::EPagesJ::API::Object::RestClient::new> - ref.hash
#                              <li>content - optional. content data - ref.hash or arrayref
#                              <li>authenticated - optional. determines if request requires authorization header (default: true) - boolean
#                          </ul> | ref.hash
#----------------------------------------------------------------------------------------
# return       $HttpResponse | response as object | HTTP::Response
#========================================================================================
sub _sendRequestWithRefreshWrapper {
    my $self = shift;
    my %hOptions = @_;

    my $MethodName = (caller(0))[3];

    # Unique ID to correlate request/response logs
    my $CID = GenerateCorrelationID();
    $hOptions{'correlationid'} = $CID;

    my ($HttpResponse, $hLog, $Error);
    my $LinkHeader;
    my $FullPagedContent = [];
    my $RetryCount = 0;
    while ((!defined($HttpResponse)
                # Include "equal" comparison since first request doesn't count as retry
                && $RetryCount <= MAX_RETRIES)
            || defined $LinkHeader) {
        eval {
            $HttpResponse = $self->_sendRequest(%hOptions);

            # If 'Link' header is present it will contain the URL to retrieve the next "page"
            # i.e: <https://inventory.izettle.com/v3/stock/8f59b9ee-828b-11eb-b512-f9a20a1c5cc2?cursor=6163cfaa-ab14-18bd-2f10-b976395f6873&limit=100>; rel="next"
            $LinkHeader = $HttpResponse->header('Link');
            if (defined $LinkHeader && $LinkHeader =~ /^(\s*<([^>]*)>\s*)/) {
                # Retrieve next page URL from second pattern in regex (first pattern is the whole expression)
                my $NextPage = $2;
                # Update request URL
                $hOptions{'url'} = $NextPage;
            }
            if (defined $LinkHeader || scalar @{$FullPagedContent}) {
                # Combine contents of each page (decoded content shall be a list of elements)
                my $aDecodedContent = $self->decodeResponseContent($HttpResponse);
                push(@{$FullPagedContent}, @{$aDecodedContent}) if scalar @{$aDecodedContent};
            }

            # See https://developer.zettle.com/docs/api/inventory-v3/overview#pagination
            # Though Zettle DOC doesn't clarify, we safely assume value is number of seconds (https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Retry-After#syntax)
            my $RetryAfterHeader = $HttpResponse->header('Retry-After');
            if (defined $RetryAfterHeader) {
                sleep($RetryAfterHeader);
            }
        };
        if (ExistsError()) {
            $Error = GetError();
            $RetryCount++;

            my $Logger = $self->_getLogger();
            my $HttpMethod = $hOptions{'method'};
            my $RequestUrl = $hOptions{'url'};
            # Refresh access token if expired and resend request
            if ($self->_isAccessTokenExpired($Error) && !$self->_isRefreshTokenInvalid($Error)) {
                my $hHeaders = $hOptions{'headers'} // {};
                $hLog = { 'Method' => $HttpMethod, 'URL' => $RequestUrl, 'Headers' => $hHeaders };
                $Logger->logCommunication("$MethodName refreshing expired access token for (HTTP Request, CID=$CID)", $hLog);
                # Refresh access token (also replaces old token in ZettleEnvironment)
                my $ForceRefresh = 1;
                my $ValidToken = $self->_getValidToken($ForceRefresh);
                $hOptions{'headers'} //= {};
                $hOptions{'headers'}->{'Authorization'} = "Bearer $ValidToken";

                $HttpResponse =  $self->_sendRequest(%hOptions);
            }
            # Handle throttling when API rate limit is reached (HTTP error 429)
            elsif ($Error->code() eq REQUEST_FAILED_CODE
                        && $Error->vars()->{'Status'} == HTTP_TOO_MANY_REQUESTS) {
                my $DelayMS = GetAPICallRetryDelay($hOptions{'url'}, $RetryCount);
                my $LogMsg = "$MethodName - RateLimitExceeded (HTTP Request, CID=$CID): "
                                ."waiting for $DelayMS milliseconds "
                                ."(Retry: ".$RetryCount." of ".MAX_RETRIES.")";
                $hLog = { 'Method' => $HttpMethod, 'URL' => $RequestUrl };
                $Logger->logCommunication($LogMsg, $hLog) unless $RetryCount > MAX_RETRIES;
                usleep($DelayMS*1000);
                $HttpResponse = undef;
            }
            # Forward all other errors
            else {
                $self->_handleErrors($Error);
            }
        } else {
            # Reset retry counter since we received a succesful response
            $RetryCount = 0;
        }
    }

    if (not defined $HttpResponse) {
        if ($RetryCount > MAX_RETRIES) {
            Error('RetryLimitReached', {
                'HttpMethod' => $hOptions{'method'},
                'RequestUrl' => $hOptions{'url'},
                'CorrelationID' => $hOptions{'correlationid'},
                'RetryCount' => $RetryCount-1
            });
        } else {
            Error($Error);
        }
    }

    # Check whether we need to merge back contents from paged responses
    if (scalar @{$FullPagedContent}) {
        # Re-encode full contents back into HttpResponse object (otherwise caller would break)
        my $EncodedContent = encode_json($FullPagedContent);
        $HttpResponse->content($EncodedContent);
    }

    return $HttpResponse;
}

#========================================================================================
# function     _sendRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->_sendRequest($hOptions);
#----------------------------------------------------------------------------------------
# description  sends an HTTP request
#----------------------------------------------------------------------------------------
# throws       JsonParserError, WebserviceError, HttpEncodingError
#----------------------------------------------------------------------------------------
# input        $hOptions | arguments and options used to construct HTTP request with keys:
#                          <ul>
#                              <li>method - HTTP method: get/put/post/delete - string
#                              <li>url - target endpoint URL - string
#                              <li>headers - HTTP header data see <function DE_EPAGES::EPagesJ::API::Object::RestClient::new> - ref.hash
#                              <li>content - optional. content data - ref.hash or arrayref
#                              <li>authenticated - optional. determines if request requires authorization header (default: true) - boolean
#                              <li>correlationid - optional. 40-character string used for request/response correlation (see <function DE_EPAGES::Zettle::API::Tools::GenerateCorrelationID) - string
#                          </ul> | ref.hash
#----------------------------------------------------------------------------------------
# return       $HttpResponse | response as object | HTTP::Response
#========================================================================================
sub _sendRequest {
    my $self = shift;
    my %hOptions = @_;
    my $MethodName = (caller(0))[3];

    # Unique ID to correlate request/response logs
    my $CorrelationID = $hOptions{'correlationid'} // GenerateCorrelationID();

    my $HttpMethod = $hOptions{'method'};
    my $Url = $hOptions{'url'};
    my $hHeaders = $hOptions{'headers'} // {};
    my $hContent = $hOptions{'content'};
    # All requests are authenticated (include Bearer token) by default, unless otherwise specified
    my $Authenticated = $hOptions{'authenticated'} // 1;

    # Add Zettle partner ID to headers
    $self->_addPartnerHeader($hHeaders);
    # Add bearer authorization to headers if required
    $self->_addAuthorization($hHeaders) if $Authenticated;

    my $Logger = $self->_getLogger();
    my $hLog = { 'Method' => $HttpMethod, 'URL' => $Url, 'Headers' => $hHeaders };
    $hLog->{'Content'} = $hContent if defined $hContent;
    $Logger->logCommunication("$MethodName (HTTP Request, CID=$CorrelationID)", $hLog);

    my $RequestUrl = URI::URL->new($Url);

    my $Response;
    # Underlying client: DE_EPAGES::EPagesJ::API::Object::HTTPClient
    # Any errors occured will be handled by caller method: _sendRequestWithRefreshWrapper
    my $RestClient = $self->_getRestClient();
    if($HttpMethod eq HTTP_GET){
        $RequestUrl->query_form($hContent) if defined $hContent;
        $Response = $RestClient->client()->get($RequestUrl, $hHeaders);
    }elsif($HttpMethod eq HTTP_POST){
        my $EncodedRequest = $self->_encodeContent($hContent, $hHeaders->{'Content-type'});
        $Response = $RestClient->client()->post($RequestUrl, $hHeaders, $EncodedRequest);
    }elsif($HttpMethod eq HTTP_PUT){
        my $EncodedRequest = $self->_encodeContent($hContent, $hHeaders->{'Content-type'});
        $Response = $RestClient->client()->put($RequestUrl, $hHeaders, $EncodedRequest);
    }elsif($HttpMethod eq HTTP_DELETE){
        $RequestUrl->query_form($hContent) if defined $hContent;
        $Response = $RestClient->client()->delete($RequestUrl, $hHeaders);
    }
    $Logger->logCommunication("$MethodName (HTTP Response, CID=$CorrelationID)", $Response);
    return $Response;
}

#========================================================================================
# function     _handleErrors
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_handleErrors($Error);
#----------------------------------------------------------------------------------------
# description  handle the errors which can happen
#----------------------------------------------------------------------------------------
# throws       JsonParserError, WebserviceError, NativeDie and all other Errors
#----------------------------------------------------------------------------------------
# input        $Error | epages error object | DE_EPAGES::Core::API::Error
#========================================================================================
sub _handleErrors {
    my $self = shift;
    my ($Error) = @_;
    my $MethodName = (caller(0))[3];

    if ($Error->code() eq 'NativeDie') {
        my $Message = $Error->message();
        my $ErrorMessage_for_encodeRequest = 'hash- or arrayref expected';
        my $ErrorMessage_for_decodeResponse = 'malformed JSON string';
        if ($Message =~ /$ErrorMessage_for_encodeRequest|$ErrorMessage_for_decodeResponse/g) {
            Error('JsonParserError', { 'Message' => $Message });
        }
        else {
            Error($Error); #forward all other NativeDies
        }
    }
    # HTTP request failed (see DE_EPAGES::EPagesJ::API::Object::HTTPClient::send())
    elsif ($Error->code() eq REQUEST_FAILED_CODE) {
        my $ErrorVars = $Error->vars();
        my $Response = HTTP::Response->new($ErrorVars->{'Status'});
        $Response->content($ErrorVars->{'Response'});

        my $Logger = $self->_getLogger();
        $Logger->logCommunication($MethodName, $ErrorVars);
        Error(WEB_SERVICE_ERROR_CODE, {
            'StatusLine' => $ErrorVars->{'StatusLine'},
            'StatusCode' => $ErrorVars->{'Status'},
            'Response' => $Response->decoded_content
        });
    }
    else {
        Error($Error); #forward all other Errors
    }

    return;
}

#========================================================================================
# function     _isAccessTokenExpired
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_isAccessTokenExpired($Error);
#----------------------------------------------------------------------------------------
# description  check error and verify if it was raised by an expired access token sent to Zettle.
#               See https://github.com/iZettle/api-documentation/blob/master/authorization.md#errors-1
#               This check is different from <function DE_EPAGES::Zettle::API::TokenHelper:isTokenExpired>
#               which checks the expiration based on the ExpiryDate stored when AccessToken is retrieved.
#----------------------------------------------------------------------------------------
# input        $Error | epages error object | DE_EPAGES::Core::API::Error
#----------------------------------------------------------------------------------------
# throws       JsonParserError, WebserviceError, NativeDie and all other Errors
#----------------------------------------------------------------------------------------
# input        $IsExpired | TRUE if error is due to an expired access token | boolean
#========================================================================================
sub _isAccessTokenExpired {
    my $self = shift;
    my ($Error)  = @_;

    my $isTokenExpired = 0;
    if (defined($Error) && ($Error->code eq WEB_SERVICE_ERROR_CODE)) {
        # ErrorVars: StatusCode (int), StatusLine (string), Response (JSON string)
        my $ErrorVars = $Error->vars();
        my $hResponse = decode_json($ErrorVars->{'Response'});

        # Example: HTTP 401 Response = {"developerMessage":"Access token expired at: NumericDate{1633598714 -> Oct 7, 2021, 9:25:14 AM UTC}","errorType":"ACCESS_TOKEN_EXPIRED"}
        $isTokenExpired = (defined $hResponse && defined $hResponse->{'errorType'} && $hResponse->{'errorType'} eq ACCESS_TOKEN_EXPIRED_CODE);
    }
    return $isTokenExpired;
}

#========================================================================================
# function     _isRefreshTokenInvalid
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_isRefreshTokenInvalid($Error);
#----------------------------------------------------------------------------------------
# description  check error and verify if it was raised by an invalid refresh token sent to Zettle.
#               See https://github.com/iZettle/api-documentation/blob/master/authorization.md#errors
#----------------------------------------------------------------------------------------
# input        $Error | epages error object | DE_EPAGES::Core::API::Error
#----------------------------------------------------------------------------------------
# throws       JsonParserError, WebserviceError, NativeDie and all other Errors
#----------------------------------------------------------------------------------------
# input        $IsExpired | TRUE if error is due to an invalid refresh token | boolean
#========================================================================================
sub _isRefreshTokenInvalid {
    my $self = shift;
    my ($Error)  = @_;

    my $isTokenInvalid = 0;
    if (defined($Error) && ($Error->code eq WEB_SERVICE_ERROR_CODE)) {
        # ErrorVars: StatusCode (int), StatusLine (string), Response (JSON string)
        my $ErrorVars = $Error->vars();
        my $hResponse = decode_json($ErrorVars->{'Response'});

        # Example: HTTP 400 Response = {"error":"invalid_grant","error_description":"invalid refresh_token"}
        $isTokenInvalid = (defined $hResponse && defined $hResponse->{'error'} && $hResponse->{'error'} eq INVALID_GRANT_CODE
                            && defined $hResponse->{'error_description'} && $hResponse->{'error_description'} eq INVALID_REFRESH_TOKEN_CODE);
    }
    return $isTokenInvalid;
}


#========================================================================================
sub _getRestClient {
    my $self = shift;
    my ($hSettings) = @_;
    $hSettings //= {};

    $self->{'RestClient'} //= DE_EPAGES::EPagesJ::API::Object::RestClient->create($self->getEnvironment(), $hSettings);
    # Update user-agent
    my $UserAgent = $self->{'RestClient'}->client()->userAgent();
    $UserAgent->agent('ePages/6 ' . $UserAgent->agent()) unless $UserAgent->agent() =~ m/ePages/i;

    return $self->{'RestClient'};
}

#========================================================================================
sub _getLogger {
    my $self = shift;

    $self->{'Logger'} //= DE_EPAGES::Zettle::API::Logger->new();
    return $self->{'Logger'};
}

#========================================================================================
# function     _addPartnerHeader
# state        private
#========================================================================================
# syntax       $self->_addPartnerHeader($hHeaders);
#----------------------------------------------------------------------------------------
# description  Add X-iZettle-Application-Id header.
#----------------------------------------------------------------------------------------
# input        $hHeaders | Request headers | ref.hash
#========================================================================================
sub _addPartnerHeader {
    my $self = shift;
    my ($hHeaders) = @_;
    
    $hHeaders->{'X-iZettle-Application-Id'} = $self->_getZettleApplicationId();

    return;
}

#========================================================================================
# function     _addAuthorization
# state        private
#========================================================================================
# syntax       $self->_addAuthorization($hHeaders);
#----------------------------------------------------------------------------------------
# description  Add valid authorization header: Bearer token.
#----------------------------------------------------------------------------------------
# input        $hHeaders | Request headers | ref.hash
#========================================================================================
sub _addAuthorization {
    my $self = shift;
    my ($hHeaders) = @_;

    # Get access token (refresh if expired).
    my $ValidToken = $self->_getValidToken();

    $hHeaders->{'Authorization'} = "Bearer $ValidToken";

    return;
}

#========================================================================================
# function     _getValidToken
# state        private
#========================================================================================
# syntax       $AccessToken = $self->_getValidToken($ForceRefresh);
#----------------------------------------------------------------------------------------
# description  Retrieve a valid access token to be used in a Zettle API call, refreshing tokens if access token is expired.
#----------------------------------------------------------------------------------------
# input        $ForceRefresh | If TRUE a refresh of the access token will be forced,
#                               even when ExpiryDate is not reached (optional. default: false) | boolean
#----------------------------------------------------------------------------------------
# return       $AccessToken | valid access token | string
#========================================================================================
sub _getValidToken {
    my $self = shift;
    my ($ForceRefresh) = @_;

    my $Environment = $self->getEnvironment();
    my $TokenHelper = DE_EPAGES::Zettle::API::TokenHelper->new($Environment);
    if($TokenHelper->isTokenExpired() || $ForceRefresh){
        my $RefreshToken = $Environment->refreshToken();
        my $AuthService = DE_EPAGES::Zettle::API::WebService::AuthService->new($Environment);
        #TODO: handle invalid refresh token
        my $hTokenInfo = $AuthService->refreshTokenData($RefreshToken);
        $TokenHelper->updateTokenData($hTokenInfo);
    }

    return $Environment->accessToken();
}

#========================================================================================
sub _getZettleApplicationId {
    my $self = shift;

    return LoadRootObject()->get('Zettle_AppClientID');
}

1;
