#========================================================================================
# package      DE_EPAGES::Zettle::API::ZettleUrlBuilder
# state        public
#----------------------------------------------------------------------------------------
# description  Build return URL for Backlink
#========================================================================================
package DE_EPAGES::Zettle::API::ZettleUrlBuilder;

use strict;

use URI;
use URI::URL;

use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Core::API::Url qw ( ParseQueryString );
use DE_EPAGES::Object::API::Factory qw( LoadRootObject );
use DE_EPAGES::Shop::API::Url qw ( BuildShopUrl );
use DE_EPAGES::Zettle::API::Constants qw ( APP_SCOPES VALIDATION_TOKEN_KEY );
use DE_EPAGES::Zettle::API::Tools qw ( GenerateCallbackToken );
use DE_EPAGES::Zettle::API::WebService::EndpointHandler qw ( GetAuthApiUrl ); ## no critic qw(RequireCorrectImports)

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Zettle::API::ZettleUrlBuilder->new();
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Zettle::API::ZettleUrlBuilder
#========================================================================================
sub new {
    my $class = shift;
    my $self  = bless {}, $class;
    return $self;
}

#========================================================================================
# function     buildAdminUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->buildAdminUrl();
#----------------------------------------------------------------------------------------
# description  Retrieve the administration URL to access Zettle merchant account.
#----------------------------------------------------------------------------------------
# return       Zettle administration URL | string
#========================================================================================
sub buildAdminUrl {
    return shift->_loadSystemAttribute('Zettle_AdminURL');
}

#========================================================================================
# function     buildConnectUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->buildConnectUrl($Servlet, $Shop);
#----------------------------------------------------------------------------------------
# description  Build the connect URL used for on-boarding to Zettle
#----------------------------------------------------------------------------------------
# input        $Servlet | a Servlet object  | DE_EPAGES::Presentation::Servlet::BasePermission
# input        $Shop | a Shop object  | DE_EPAGES::Shop::API::Object::Shop
#----------------------------------------------------------------------------------------
# return       $URL->as_string() | Zettle connect url | string
#========================================================================================
sub buildConnectUrl {
    my $self = shift;
    my ($Servlet, $Shop) = @_;

    my $StateURL = $self->_buildStateURL($Servlet, $Shop);

    my $hQueryParameters = $self->_buildQueryParameters($StateURL);

    my $url = $self->_buildEncodedOnboardingUrl($hQueryParameters);

    return $url;
}

#========================================================================================
# function     buildRedirectToMboUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->buildRedirectToMboUrl( $hQueryParameters );
#----------------------------------------------------------------------------------------
# description  Build redirect to MBO URL using returned data from Zettle
#----------------------------------------------------------------------------------------
# input        $hQueryParameters | Parameters as returned from Zettle with keys
#               <ul>
#                   <li>code
#                   <li>state
#               </ul> | ref.hash
#----------------------------------------------------------------------------------------
# throws       NoBacklinkFound   If $hQueryParameters does not contain 'state' parameter
#----------------------------------------------------------------------------------------
# return       $URL->as_string() | BacklinkUrl with AuthorizationCode or Error | string
#========================================================================================
sub buildRedirectToMboUrl {
    my $self = shift;
    my ($hQueryParameters) = @_;
    my $aBacklinkUrl = $hQueryParameters->{'state'};

    if ( (!isValid($aBacklinkUrl)) or (!isValid($aBacklinkUrl->[0])) ) {
        Error('NoURLFound');
    }

    return $self->_buildRedirectUrl($hQueryParameters);
}

#========================================================================================
# function     buildBacklinkRedirectUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->buildBacklinkRedirectUrl();
#----------------------------------------------------------------------------------------
# description  Build backlink from 'Zettle_AppRedirectURL' system attribute (if present) or build from current domain (fallback)
# example      https://<DOMAIN>/epages/Site.admin?ChangeAction=BacklinkZettleOnboarding)
#----------------------------------------------------------------------------------------
# return       Retrieve backlink used by Zettle to return information to the shop (URL to be sent in 'state' parameter during OAuth) | string
#========================================================================================
sub buildBacklinkRedirectUrl {
    my $self = shift;

    # 1) Load redirect URL from system attribute (if present)
    my $RedirectUrl = $self->_loadSystemAttribute('Zettle_AppRedirectURL');
    return $RedirectUrl if isValid($RedirectUrl);
    
    # 2) Fallback: build redirect URL from current domain
    $RedirectUrl = $self->_loadSystemAttribute('ProtocolAndServerSSL');
    $RedirectUrl .= '/epages/Site.admin?ChangeAction=BacklinkZettleOnboarding';

    return $RedirectUrl;
}

# Return callback URL to handle InventoryBalanceChanged event from Zettle Webhooks
sub buildWebhookStockUrl {
    my $self = shift;
    my ($ZettleEnvironment) = @_;

    my $hOptions = {
        'ChangeAction' => 'ZettleStockCallback',
        'UseObjectPath' => 1,
        # webhooks are not session-dependent so we don't include security token
        'AddToken' => 0
    };

    return $self->_buildAdminUrl($ZettleEnvironment, $hOptions);
}

#========================================================================================

# Retrieve MBO URL for backlink 'state' parameter, including ChangeAction and ObjectID
# Sample: https://<SERVER>/epages/<SHOPNAME>.admin?ObjectID=<SHOP_ID>&ChangeAction=ZettleOAuthCallback
sub _buildStateURL {
    my $self = shift;
    my ($Servlet, $Shop) = @_;

    my $hOptions = {
        # Action that will handle backlink and process auth code for OAuth flow
        'ChangeAction' => 'ZettleOAuthCallback'
    };

    return $self->_buildAdminUrl($Shop, $hOptions, $Servlet);
}

#========================================================================================
# function      _buildAdminUrl
# state        private
#----------------------------------------------------------------------------------------
# syntax       $URL = $self->_buildAdminUrl($Object, $hOptions, $Servlet)
#               $URL = $self->_buildAdminUrl($Object, $hOptions)
#----------------------------------------------------------------------------------------
# description  Create an MBO URL.
#----------------------------------------------------------------------------------------
# input        $Object | any object that belongs to a shop | object
# input        $hOptions | action information with either ViewAction or ChangeAction present:
#                          <ul>
#                              <li>ViewAction    - Name(s) of ViewAction(s) - string or array of strings
#                              <li>ChangeAction  - Name(s) of ChangeAction(s) - string or array of strings
#                              <li>AddToken      - add security token or not (optional, defaults to 1) - boolean
#                              <li>UseObjectPath - use object path rather than objectid (optional, defaults to 0) - boolean
#                          </ul> | ref.hash
# input        $Servlet | a Servlet object (optional, but required unless AddToken option is false)| DE_EPAGES:Presentation::Servlet::BasePermission
#----------------------------------------------------------------------------------------
# return       $URL | Admin URL | string
#========================================================================================
sub _buildAdminUrl {
    my $self = shift;
    my ($Object, $hOptions, $Servlet) = @_;

    my $hParams = {};
    foreach my $ActionType ( qw( ViewAction ChangeAction ) ){
        my $OptionValue = $hOptions->{$ActionType};
        $hParams->{$ActionType} = $OptionValue if defined $OptionValue;
    }

    my $hBuildOptions = {
        'Type' => 'admin',
        'UseSSL' => 1,
        'AddToken' => 1,
    };
    my $AddToken = $hOptions->{'AddToken'};
    $hBuildOptions->{'AddToken'} = 0 if (defined $AddToken && $AddToken == 0);
    $hBuildOptions->{'UseObjectPath'} = 1 if $hOptions->{'UseObjectPath'};

    my $AdminUrl = BuildShopUrl($Object, $hParams, $hBuildOptions, $Servlet);
    # Generate callback token and add it to admin URL
    my $CallbackToken = GenerateCallbackToken($Object);
    
    return sprintf('%s&'.VALIDATION_TOKEN_KEY.'=%s', $AdminUrl, $CallbackToken);
}

# Retrieve a hash with all query parameters required for backlink
sub _buildQueryParameters {
    my $self = shift;
    my ($stateURL) = @_;

    my $ClientID = $self->_loadSystemAttribute('Zettle_AppClientID');
    my $redirectUrl = $self->buildBacklinkRedirectUrl();

    my $hQueryParameters = {
        'response_type'   => 'code',
        'client_id'       => $ClientID,
        'scope'           => join(' ', APP_SCOPES),
        'redirect_uri'    => $redirectUrl,
        'state'           => $stateURL,
    };

    # Add UTM parameters to connect URL
    my $UtmParameters = $self->_getUtmParameters();

    return { %$hQueryParameters, %$UtmParameters };
}

# Encode URL with parameters
sub _buildEncodedOnboardingUrl {
    my $self = shift;
    my ($hQueryParameters) = @_;

    my $ZettleOAuthURL = GetAuthApiUrl();

    my $Url = URI::URL->new($ZettleOAuthURL);
    $Url->query_form($hQueryParameters);
    my $hBacklinkUrlParameters = ParseQueryString($Url->equery(), 'utf-8');
    $Url->query_form( $hBacklinkUrlParameters );

    return $Url;
}

# Build redirect URL from data returned by Zettle Onboarding process
sub _buildRedirectUrl {
    my $self = shift;
    my ($hQueryParameters) = @_;

    # Original Change action must be removed
    delete($hQueryParameters->{'ChangeAction'});

    # State contains the URL which is backlink to MBO. No need to send it with QueryString again
    my $BackLinkUrl = delete($hQueryParameters->{'state'});
    my $URL = URI->new($BackLinkUrl->[0]);
    my $hBacklinkUrlParameters = ParseQueryString($URL->query(), 'utf-8');
    
    my $hQueryParam = {%{$hBacklinkUrlParameters}, %{$hQueryParameters}};
    $URL->query_form($hQueryParam);

    return $URL->as_string();
}

sub _getUtmParameters {
    my $self = shift;

    my $UtmQueryString = $self->_loadSystemAttribute('Zettle_UTMParameters');
    $UtmQueryString = substr($UtmQueryString, 1) if index($UtmQueryString, '?') >= 0;

    my $UtmParameters = {};
    my @aUtmParams = split(/&/, $UtmQueryString);
    foreach my $UtmParam (@aUtmParams){
        my @Parts = split(/=/, $UtmParam);
        $UtmParameters->{$Parts[0]} = $Parts[1];
    }

    return $UtmParameters;
}

# Return the value of the given system attribute
sub _loadSystemAttribute {
    my $self = shift;
    my ($AttributeName) = @_;

    my $System = LoadRootObject();
    my $AttributeValue = $System->get($AttributeName);

    return $AttributeValue;
}

1;
